# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import copy

class BacktrackParameter:

	__slots__ = (
		"circular_dependency",
		"needed_unstable_keywords", "runtime_pkg_mask", "needed_use_config_changes", "needed_license_changes",
		"prune_rebuilds", "rebuild_list", "reinstall_list", "needed_p_mask_changes",
		"slot_operator_mask_built", "slot_operator_replace_installed"
	)

	def __init__(self):
		self.circular_dependency = {}
		self.needed_unstable_keywords = set()
		self.needed_p_mask_changes = set()
		self.runtime_pkg_mask = {}
		self.needed_use_config_changes = {}
		self.needed_license_changes = {}
		self.rebuild_list = set()
		self.reinstall_list = set()
		self.slot_operator_replace_installed = set()
		self.slot_operator_mask_built = set()
		self.prune_rebuilds = False

	def __deepcopy__(self, memo=None):
		if memo is None:
			memo = {}
		result = BacktrackParameter()
		memo[id(self)] = result

		#Shallow copies are enough here, as we only need to ensure that nobody adds stuff
		#to our sets and dicts. The existing content is immutable.
		result.circular_dependency = copy.copy(self.circular_dependency)
		result.needed_unstable_keywords = copy.copy(self.needed_unstable_keywords)
		result.needed_p_mask_changes = copy.copy(self.needed_p_mask_changes)
		result.needed_use_config_changes = copy.copy(self.needed_use_config_changes)
		result.needed_license_changes = copy.copy(self.needed_license_changes)
		result.rebuild_list = copy.copy(self.rebuild_list)
		result.reinstall_list = copy.copy(self.reinstall_list)
		result.slot_operator_replace_installed = copy.copy(self.slot_operator_replace_installed)
		result.slot_operator_mask_built = self.slot_operator_mask_built.copy()
		result.prune_rebuilds = self.prune_rebuilds

		# runtime_pkg_mask contains nested dicts that must also be copied
		result.runtime_pkg_mask = {}
		for k, v in self.runtime_pkg_mask.items():
			result.runtime_pkg_mask[k] = copy.copy(v)

		return result

	def __eq__(self, other):
		return self.circular_dependency == other.circular_dependency and \
			self.needed_unstable_keywords == other.needed_unstable_keywords and \
			self.needed_p_mask_changes == other.needed_p_mask_changes and \
			self.runtime_pkg_mask == other.runtime_pkg_mask and \
			self.needed_use_config_changes == other.needed_use_config_changes and \
			self.needed_license_changes == other.needed_license_changes and \
			self.rebuild_list == other.rebuild_list and \
			self.reinstall_list == other.reinstall_list and \
			self.slot_operator_replace_installed == other.slot_operator_replace_installed and \
			self.slot_operator_mask_built == other.slot_operator_mask_built and \
			self.prune_rebuilds == other.prune_rebuilds


class _BacktrackNode:

	__slots__ = (
		"parameter", "depth", "mask_steps", "terminal",
	)

	def __init__(self, parameter=BacktrackParameter(), depth=0, mask_steps=0, terminal=True):
		self.parameter = parameter
		self.depth = depth
		self.mask_steps = mask_steps
		self.terminal = terminal

	def __eq__(self, other):
		return self.parameter == other.parameter


class Backtracker:

	__slots__ = (
		"_max_depth", "_unexplored_nodes", "_current_node", "_nodes", "_root",
	)

	def __init__(self, max_depth):
		self._max_depth = max_depth
		self._unexplored_nodes = []
		self._current_node = None
		self._nodes = []

		self._root = _BacktrackNode()
		self._add(self._root)


	def _add(self, node, explore=True):
		"""
		Adds a newly computed backtrack parameter. Makes sure that it doesn't already exist and
		that we don't backtrack deeper than we are allowed by --backtrack.
		"""
		if not self._check_runtime_pkg_mask(node.parameter.runtime_pkg_mask):
			return

		if node.mask_steps <= self._max_depth and node not in self._nodes:
			if explore:
				self._unexplored_nodes.append(node)
			self._nodes.append(node)


	def get(self):
		"""
		Returns a backtrack parameter. The backtrack graph is explored with depth first.
		"""
		if self._unexplored_nodes:
			node = self._unexplored_nodes.pop()
			self._current_node = node
			return copy.deepcopy(node.parameter)
		return None


	def __len__(self):
		return len(self._unexplored_nodes)

	def _check_runtime_pkg_mask(self, runtime_pkg_mask):
		"""
		If a package gets masked that caused other packages to be masked
		before, we revert the mask for other packages (bug 375573).
		"""

		for pkg, mask_info in runtime_pkg_mask.items():

			if "missing dependency" in mask_info or \
				"slot_operator_mask_built" in mask_info:
				continue

			entry_is_valid = False
			any_conflict_parents = False

			for ppkg, patom in runtime_pkg_mask[pkg].get("slot conflict", set()):
				any_conflict_parents = True
				if ppkg not in runtime_pkg_mask:
					entry_is_valid = True
					break
			else:
				if not any_conflict_parents:
					# Even though pkg was involved in a slot conflict
					# where it was matched by all involved parent atoms,
					# consider masking it in order to avoid a missed
					# update as in bug 692746.
					entry_is_valid = True

			if not entry_is_valid:
				return False

		return True

	def _feedback_slot_conflicts(self, conflicts_data):
		# Only create BacktrackNode instances for the first
		# conflict which occurred, since the conflicts that
		# occurred later may have been caused by the first
		# conflict.
		self._feedback_slot_conflict(conflicts_data[0])

	def _feedback_slot_conflict(self, conflict_data):
		for similar_pkgs in conflict_data:
			new_node = copy.deepcopy(self._current_node)
			new_node.depth += 1
			new_node.mask_steps += 1
			new_node.terminal = False
			for pkg, parent_atoms in similar_pkgs:
				new_node.parameter.runtime_pkg_mask.setdefault(
					pkg, {})["slot conflict"] = parent_atoms
			self._add(new_node)


	def _feedback_missing_dep(self, dep):
		new_node = copy.deepcopy(self._current_node)
		new_node.depth += 1
		new_node.mask_steps += 1
		new_node.terminal = False

		new_node.parameter.runtime_pkg_mask.setdefault(
			dep.parent, {})["missing dependency"] = \
				set([(dep.parent, dep.root, dep.atom)])

		self._add(new_node)


	def _feedback_config(self, changes, explore=True):
		"""
		Handle config changes. Don't count config changes for the maximum backtrack depth.
		"""
		new_node = copy.deepcopy(self._current_node)
		new_node.depth += 1
		para = new_node.parameter

		for change, data in changes.items():
			if change == "circular_dependency":
				for pkg, circular_children in data.items():
					para.circular_dependency.setdefault(pkg, set()).update(circular_children)
			elif change == "needed_unstable_keywords":
				para.needed_unstable_keywords.update(data)
			elif change == "needed_p_mask_changes":
				para.needed_p_mask_changes.update(data)
			elif change == "needed_license_changes":
				for pkg, missing_licenses in data:
					para.needed_license_changes.setdefault(pkg, set()).update(missing_licenses)
			elif change == "needed_use_config_changes":
				for pkg, (new_use, new_changes) in data:
					para.needed_use_config_changes[pkg] = (new_use, new_changes)
			elif change == "slot_conflict_abi":
				new_node.terminal = False
			elif change == "slot_operator_mask_built":
				para.slot_operator_mask_built.update(data)
				for pkg, mask_reasons in data.items():
					para.runtime_pkg_mask.setdefault(pkg,
						{}).update(mask_reasons)
			elif change == "slot_operator_replace_installed":
				para.slot_operator_replace_installed.update(data)
			elif change == "rebuild_list":
				para.rebuild_list.update(data)
			elif change == "reinstall_list":
				para.reinstall_list.update(data)
			elif change == "prune_rebuilds":
				para.prune_rebuilds = True
				para.slot_operator_replace_installed.clear()
				for pkg in para.slot_operator_mask_built:
					runtime_masks = para.runtime_pkg_mask.get(pkg)
					if runtime_masks is None:
						continue
					runtime_masks.pop("slot_operator_mask_built", None)
					if not runtime_masks:
						para.runtime_pkg_mask.pop(pkg)
				para.slot_operator_mask_built.clear()

		self._add(new_node, explore=explore)
		self._current_node = new_node


	def feedback(self, infos):
		"""
		Takes information from the depgraph and computes new backtrack parameters to try.
		"""
		assert self._current_node is not None, "call feedback() only after get() was called"

		#Not all config changes require a restart, that's why they can appear together
		#with other conflicts.
		if "config" in infos:
			self._feedback_config(infos["config"], explore=(len(infos)==1))

		#There is at most one of the following types of conflicts for a given restart.
		if "slot conflict" in infos:
			self._feedback_slot_conflicts(infos["slot conflict"])
		elif "missing dependency" in infos:
			self._feedback_missing_dep(infos["missing dependency"])


	def backtracked(self):
		"""
		If we didn't backtrack, there is only the root.
		"""
		return len(self._nodes) > 1


	def get_best_run(self):
		"""
		Like, get() but returns the backtrack parameter that has as many config changes as possible,
		but has no masks. This makes --autounmask effective, but prevents confusing error messages
		with "masked by backtracking".
		"""
		best_node = self._root
		for node in self._nodes:
			if node.terminal and node.depth > best_node.depth:
				best_node = node

		return copy.deepcopy(best_node.parameter)



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.util.SlotObject import SlotObject
from _emerge.DepPriority import DepPriority

class Dependency(SlotObject):
	__slots__ = ("atom", "blocker", "child", "depth",
		"parent", "onlydeps", "priority", "root", "want_update",
		"collapsed_parent", "collapsed_priority")
	def __init__(self, **kwargs):
		SlotObject.__init__(self, **kwargs)
		if self.priority is None:
			self.priority = DepPriority()
		if self.depth is None:
			self.depth = 0
		if self.collapsed_parent is None:
			self.collapsed_parent = self.parent
		if self.collapsed_priority is None:
			self.collapsed_priority = self.priority



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess

try:
	import threading
except ImportError:
	# dummy_threading will not suffice
	threading = None

from portage import os
from portage.tests import TestCase
from portage.util._async.PopenProcess import PopenProcess
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util._async.PipeReaderBlockingIO import PipeReaderBlockingIO

class PopenPipeBlockingIOTestCase(TestCase):
	"""
	Test PopenProcess, which can be useful for Jython support:
		* use subprocess.Popen since Jython does not support os.fork()
		* use blocking IO with threads, since Jython does not support
		  fcntl non-blocking IO)
	"""

	_echo_cmd = "echo -n '%s'"

	def _testPipeReader(self, test_string):
		"""
		Use a poll loop to read data from a pipe and assert that
		the data written to the pipe is identical to the data
		read from the pipe.
		"""

		producer = PopenProcess(proc=subprocess.Popen(
			["bash", "-c", self._echo_cmd % test_string],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT),
			pipe_reader=PipeReaderBlockingIO(), scheduler=global_event_loop())

		consumer = producer.pipe_reader
		consumer.input_files = {"producer" : producer.proc.stdout}

		producer.start()
		producer.wait()

		self.assertEqual(producer.returncode, os.EX_OK)
		self.assertEqual(consumer.returncode, os.EX_OK)

		return consumer.getvalue().decode('ascii', 'replace')

	def testPopenPipeBlockingIO(self):

		if threading is None:
			skip_reason = "threading disabled"
			self.portage_skip = "threading disabled"
			self.assertFalse(True, skip_reason)
			return

		for x in (1, 2, 5, 6, 7, 8, 2**5, 2**10, 2**12, 2**13, 2**14):
			test_string = x * "a"
			output = self._testPipeReader(test_string)
			self.assertEqual(test_string, output,
				"x = %s, len(output) = %s" % (x, len(output)))



# Copyright 2018-2021 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

import portage
portage.proxy.lazyimport.lazyimport(globals(),
	'portage.util.futures:asyncio',
)

# A marker for iscoroutinefunction.
_is_coroutine = object()


def _iscoroutinefunction(func):
	"""
	Return True if func is a decorated coroutine function
	created with the coroutine decorator for this module.
	"""
	return getattr(func, '_is_coroutine', None) is _is_coroutine


def coroutine(generator_func):
	"""
	A decorator for a generator function that behaves as coroutine function.
	The generator should yield a Future instance in order to wait for it,
	and the result becomes the result of the current yield-expression,
	via the PEP 342 generator send() method.

	The decorated function returns a Future which is done when the generator
	is exhausted. The generator can return a value via the coroutine_return
	function.

	@param generator_func: A generator function that yields Futures, and
		will receive the result of each Future as the result of the
		corresponding yield-expression.
	@type generator_func: function
	@rtype: function
	@return: A function which calls the given generator function and
		returns a Future that is done when the generator is exhausted.
	"""
	# Note that functools.partial does not work for decoration of
	# methods, since it doesn't implement the descriptor protocol.
	# This problem is solve by defining a wrapper function.
	@functools.wraps(generator_func)
	def wrapped(*args, **kwargs):
		return _generator_future(generator_func, *args, **kwargs)
	wrapped._is_coroutine = _is_coroutine
	return wrapped


def coroutine_return(result=None):
	"""
	Terminate the current coroutine and set the result of the associated
	Future.

	@param result: of the current coroutine's Future
	@type object
	"""
	raise _CoroutineReturnValue(result)


def _generator_future(generator_func, *args, **kwargs):
	"""
	Call generator_func with the given arguments, and return a Future
	that is done when the resulting generation is exhausted. If a
	keyword argument named 'loop' is given, then it is used instead of
	the default event loop.
	"""
	loop = kwargs.get('loop')
	loop = asyncio._wrap_loop(loop)
	result = loop.create_future()
	_GeneratorTask(generator_func(*args, **kwargs), result, loop=loop)
	return result


class _CoroutineReturnValue(Exception):
	def __init__(self, result):
		self.result = result


class _GeneratorTask:
	"""
	Asynchronously executes the generator to completion, waiting for
	the result of each Future that it yields, and sending the result
	to the generator.
	"""
	def __init__(self, generator, result, loop):
		self._generator = generator
		self._result = result
		self._current_task = None
		self._loop = loop
		result.add_done_callback(self._cancel_callback)
		loop.call_soon(self._next)

	def _cancel_callback(self, result):
		if result.cancelled() and self._current_task is not None:
			# The done callback for self._current_task invokes
			# _next in either case here.
			self._current_task.done() or self._current_task.cancel()

	def _next(self, previous=None):
		self._current_task = None
		if self._result.cancelled():
			if previous is not None:
				# Consume exceptions, in order to avoid triggering
				# the event loop's exception handler.
				previous.cancelled() or previous.exception()

			# This will throw asyncio.CancelledError in the coroutine if
			# there's an opportunity (yield) before the generator raises
			# StopIteration.
			previous = self._result
		try:
			if previous is None:
				future = next(self._generator)
			elif previous.cancelled():
				future = self._generator.throw(asyncio.CancelledError())
			elif previous.exception() is None:
				future = self._generator.send(previous.result())
			else:
				future = self._generator.throw(previous.exception())

		except asyncio.CancelledError:
			self._result.cancel()
		except _CoroutineReturnValue as e:
			if not self._result.cancelled():
				self._result.set_result(e.result)
		except StopIteration:
			if not self._result.cancelled():
				self._result.set_result(None)
		except Exception as e:
			if not self._result.cancelled():
				self._result.set_exception(e)
		else:
			self._current_task = asyncio.ensure_future(future, loop=self._loop)
			self._current_task.add_done_callback(self._next)



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop


class FutureDoneCallbackTestCase(TestCase):

	def testFutureDoneCallback(self):

		event_loop = global_event_loop()

		def done_callback(finished):
			done_callback_called.set_result(True)

		done_callback_called = event_loop.create_future()
		finished = event_loop.create_future()
		finished.add_done_callback(done_callback)
		event_loop.call_soon(finished.set_result, True)
		event_loop.run_until_complete(done_callback_called)

		def done_callback2(finished):
			done_callback2_called.set_result(True)

		done_callback_called = event_loop.create_future()
		done_callback2_called = event_loop.create_future()
		finished = event_loop.create_future()
		finished.add_done_callback(done_callback)
		finished.add_done_callback(done_callback2)
		finished.remove_done_callback(done_callback)
		event_loop.call_soon(finished.set_result, True)
		event_loop.run_until_complete(done_callback2_called)

		self.assertFalse(done_callback_called.done())



# Copyright 2009-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.exception import InvalidDependString
from portage.dep import Atom, use_reduce

class UseReduceTestCase:
	def __init__(self, deparray, uselist=[], masklist=[],
	             matchall=0, excludeall=[], is_src_uri=False,
	             eapi='0', opconvert=False, flat=False, expected_result=None,
	             is_valid_flag=None, token_class=None, subset=None):
		self.deparray = deparray
		self.uselist = uselist
		self.masklist = masklist
		self.matchall = matchall
		self.excludeall = excludeall
		self.is_src_uri = is_src_uri
		self.eapi = eapi
		self.opconvert = opconvert
		self.flat = flat
		self.is_valid_flag = is_valid_flag
		self.token_class = token_class
		self.subset = subset
		self.expected_result = expected_result

	def run(self):
		try:
			return use_reduce(self.deparray, self.uselist, self.masklist,
				self.matchall, self.excludeall, self.is_src_uri, self.eapi,
				self.opconvert, self.flat, self.is_valid_flag, self.token_class,
				subset=self.subset)
		except InvalidDependString as e:
			raise InvalidDependString("%s: %s" % (e, self.deparray))

class UseReduce(TestCase):

	def always_true(self, ununsed_parameter):
		return True

	def always_false(self, ununsed_parameter):
		return False

	def testUseReduce(self):

		EAPI_WITH_SRC_URI_ARROWS = "2"
		EAPI_WITHOUT_SRC_URI_ARROWS = "0"

		test_cases = (
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["a", "b", "c", "d"],
				expected_result=["A", "B"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["a", "b", "c", "d"],
				subset=["b"],
				expected_result=["B"]
				),
			UseReduceTestCase(
				"bar? ( || ( foo bar? ( baz ) ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['||', ['foo', 'baz']]
				),
			UseReduceTestCase(
				"bar? ( foo bar? ( baz ) foo )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['foo', 'baz', 'foo']
				),
			UseReduceTestCase(
				"|| ( ( a b ) ( c d ) )",
				uselist=[],
				subset=["bar"],
				expected_result=[]
				),
			UseReduceTestCase(
				"|| ( ( a b ) ( bar? ( c d ) e f ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['c', 'd']
				),
			UseReduceTestCase(
				"( a b ) ( c d bar? ( e f baz? ( g h ) ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['e', 'f']
				),
			UseReduceTestCase(
				"( a b ) ( c d bar? ( e f baz? ( g h ) ) )",
				uselist=["bar", "baz"],
				subset=["bar"],
				expected_result=['e', 'f', 'g', 'h']
				),
			UseReduceTestCase(
				"( bar? ( a b ) ( bar? ( c d ) ) ) ( e f )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['a', 'b', 'c', 'd']
				),
			UseReduceTestCase(
				"|| ( foo bar? ( baz ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=["baz"]
				),
			UseReduceTestCase(
				"|| ( || ( bar? ( a || ( b c || ( d e ) ) ) ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['a', '||', ['b', 'c', 'd', 'e']]
				),
			UseReduceTestCase(
				"|| ( || ( bar? ( a || ( ( b c ) ( d e ) ) ) ) )",
				uselist=["bar"],
				subset=["bar"],
				expected_result=['a', '||', [['b', 'c'], ['d', 'e']]]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["a", "b", "c"],
				expected_result=["A", "B", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["b", "c"],
				expected_result=["B", "D"]
				),

			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				matchall=True,
				expected_result=["A", "B", "C", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				masklist=["a", "c"],
				expected_result=["C", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				matchall=True,
				masklist=["a", "c"],
				expected_result=["B", "C", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["a", "b"],
				masklist=["a", "c"],
				expected_result=["B", "C", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				excludeall=["a", "c"],
				expected_result=["D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				uselist=["b"],
				excludeall=["a", "c"],
				expected_result=["B", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				matchall=True,
				excludeall=["a", "c"],
				expected_result=["A", "B", "D"]
				),
			UseReduceTestCase(
				"a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
				matchall=True,
				excludeall=["a", "c"],
				masklist=["b"],
				expected_result=["A", "D"]
				),

			UseReduceTestCase(
				"a? ( b? ( AB ) )",
				uselist=["a", "b"],
				expected_result=["AB"]
				),
			UseReduceTestCase(
				"a? ( b? ( AB ) C )",
				uselist=["a"],
				expected_result=["C"]
				),
			UseReduceTestCase(
				"a? ( b? ( || ( AB CD ) ) )",
				uselist=["a", "b"],
				expected_result=["||", ["AB", "CD"]]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( A ) b? ( B ) ) )",
				uselist=["a", "b"],
				expected_result=["||", ["A", "B"]]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( A ) b? ( B ) ) )",
				uselist=["a"],
				expected_result=["A"]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( A ) b? ( B ) ) )",
				uselist=[],
				expected_result=[]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
				uselist=[],
				expected_result=[]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
				uselist=["a"],
				expected_result=["A"]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
				uselist=["b"],
				expected_result=["B"]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
				uselist=["c"],
				expected_result=[]
				),
			UseReduceTestCase(
				"|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
				uselist=["a", "c"],
				expected_result=["||", ["A", "C"]]
				),

			# paren_reduce tests
			UseReduceTestCase(
				"A",
				expected_result=["A"]),
			UseReduceTestCase(
				"( A )",
				expected_result=["A"]),
			UseReduceTestCase(
				"|| ( A B )",
				expected_result=["||", ["A", "B"]]),
			UseReduceTestCase(
				"|| ( ( A B ) C )",
				expected_result=["||", [["A", "B"], "C"]]),
			UseReduceTestCase(
				"|| ( ( A B ) ( C ) )",
				expected_result=["||", [["A", "B"], "C"]]),
			UseReduceTestCase(
				"|| ( A || ( B C ) )",
				expected_result=["||", ["A", "B", "C"]]),
			UseReduceTestCase(
				"|| ( A || ( B C D ) )",
				expected_result=["||", ["A", "B", "C", "D"]]),
			UseReduceTestCase(
				"|| ( A || ( B || ( C D ) E ) )",
				expected_result=["||", ["A", "B", "C", "D", "E"]]),
			UseReduceTestCase(
				"( || ( ( ( A ) B ) ) )",
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"( || ( || ( ( A ) B ) ) )",
				expected_result=["||", ["A", "B"]]),
			UseReduceTestCase(
				"( || ( || ( ( A ) B ) ) )",
				expected_result=["||", ["A", "B"]]),
			UseReduceTestCase(
				"|| ( A )",
				expected_result=["A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				expected_result=["A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				uselist=["foo"],
				expected_result=["||", ["A", "B"]]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				expected_result=[]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				uselist=["foo", "bar"],
				expected_result=["||", ["A", "B"]]),
			UseReduceTestCase(
				"A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B",
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"|| ( A ) || ( B )",
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				expected_result=[]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				uselist=["foo"],
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"|| ( A B ) C",
				expected_result=['||', ['A', 'B'], 'C']),
			UseReduceTestCase(
				"A || ( B C )",
				expected_result=['A', '||', ['B', 'C']]),

			# SRC_URI stuff
			UseReduceTestCase(
				"http://foo/bar -> blah.tbz2",
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=["http://foo/bar", "->", "blah.tbz2"]),
			UseReduceTestCase(
				"foo? ( http://foo/bar -> blah.tbz2 )",
				uselist=[],
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=[]),
			UseReduceTestCase(
				"foo? ( http://foo/bar -> blah.tbz2 )",
				uselist=["foo"],
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=["http://foo/bar", "->", "blah.tbz2"]),
			UseReduceTestCase(
				"http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )",
				uselist=[],
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=["http://foo/bar", "->", "bar.tbz2"]),
			UseReduceTestCase(
				"http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )",
				uselist=["foo"],
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=["http://foo/bar", "->", "bar.tbz2", "ftp://foo/a"]),
			UseReduceTestCase(
				"http://foo.com/foo http://foo/bar -> blah.tbz2",
				uselist=["foo"],
				is_src_uri=True,
				eapi=EAPI_WITH_SRC_URI_ARROWS,
				expected_result=["http://foo.com/foo", "http://foo/bar", "->", "blah.tbz2"]),

			# opconvert tests
			UseReduceTestCase(
				"A",
				opconvert=True,
				expected_result=["A"]),
			UseReduceTestCase(
				"( A )",
				opconvert=True,
				expected_result=["A"]),
			UseReduceTestCase(
				"|| ( A B )",
				opconvert=True,
				expected_result=[['||', 'A', 'B']]),
			UseReduceTestCase(
				"|| ( ( A B ) C )",
				opconvert=True,
				expected_result=[['||', ['A', 'B'], 'C']]),
			UseReduceTestCase(
				"|| ( A || ( B C ) )",
				opconvert=True,
				expected_result=[['||', 'A', 'B', 'C']]),
			UseReduceTestCase(
				"|| ( A || ( B C D ) )",
				opconvert=True,
				expected_result=[['||', 'A', 'B', 'C', 'D']]),
			UseReduceTestCase(
				"|| ( A || ( B || ( C D ) E ) )",
				expected_result=["||", ["A", "B", "C", "D", "E"]]),
			UseReduceTestCase(
				"( || ( ( ( A ) B ) ) )",
				opconvert=True,
				expected_result=['A', 'B']),
			UseReduceTestCase(
				"( || ( || ( ( A ) B ) ) )",
				opconvert=True,
				expected_result=[['||', 'A', 'B']]),
			UseReduceTestCase(
				"|| ( A B ) C",
				opconvert=True,
				expected_result=[['||', 'A', 'B'], 'C']),
			UseReduceTestCase(
				"A || ( B C )",
				opconvert=True,
				expected_result=['A', ['||', 'B', 'C']]),
			UseReduceTestCase(
				"A foo? ( || ( B || ( bar? ( || ( C D E ) ) !bar? ( F ) ) ) ) G",
				uselist=["foo", "bar"],
				opconvert=True,
				expected_result=['A', ['||', 'B', 'C', 'D', 'E'], 'G']),
			UseReduceTestCase(
				"A foo? ( || ( B || ( bar? ( || ( C D E ) ) !bar? ( F ) ) ) ) G",
				uselist=["foo", "bar"],
				opconvert=False,
				expected_result=['A', '||', ['B', 'C', 'D', 'E'], 'G']),

			UseReduceTestCase(
				"|| ( A )",
				opconvert=True,
				expected_result=["A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				expected_result=["A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				uselist=["foo"],
				opconvert=True,
				expected_result=[['||', 'A', 'B']]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				opconvert=True,
				expected_result=[]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				uselist=["foo", "bar"],
				opconvert=True,
				expected_result=[['||', 'A', 'B']]),
			UseReduceTestCase(
				"A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B",
				opconvert=True,
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"|| ( A ) || ( B )",
				opconvert=True,
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				opconvert=True,
				expected_result=[]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				uselist=["foo"],
				opconvert=True,
				expected_result=["A", "B"]),
			UseReduceTestCase(
				"|| ( foo? ( || ( A B ) ) )",
				uselist=["foo"],
				opconvert=True,
				expected_result=[['||', 'A', 'B']]),

			UseReduceTestCase(
				"|| ( ( A B ) foo? ( || ( C D ) ) )",
				uselist=["foo"],
				opconvert=True,
				expected_result=[['||', ['A', 'B'], 'C', 'D']]),

			UseReduceTestCase(
				"|| ( ( A B ) foo? ( || ( C D ) ) )",
				uselist=["foo"],
				opconvert=False,
				expected_result=['||', [['A', 'B'], 'C', 'D']]),

			UseReduceTestCase(
				"|| ( ( A B ) || ( C D ) )",
				expected_result=['||', [['A', 'B'], 'C', 'D']]),

			UseReduceTestCase(
				"|| ( ( A B ) || ( C D || ( E ( F G ) || ( H ) ) ) )",
				expected_result=['||', [['A', 'B'], 'C', 'D', 'E', ['F', 'G'], 'H']]),

			UseReduceTestCase(
				"|| ( ( A B ) || ( C D || ( E ( F G ) || ( H ) ) ) )",
				opconvert=True,
				expected_result=[['||', ['A', 'B'], 'C', 'D', 'E', ['F', 'G'], 'H']]),

			UseReduceTestCase(
				"|| ( foo? ( A B ) )",
				uselist=["foo"],
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( || ( foo? ( A B ) ) )",
				uselist=["foo"],
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( || ( || ( a? ( b? ( c? ( || ( || ( || ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) )",
				uselist=["a", "b", "c", "d", "e", "f"],
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( || ( ( || ( a? ( ( b? ( c? ( || ( || ( || ( ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) ) ) ) )",
				uselist=["a", "b", "c", "d", "e", "f"],
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( ( A ( || ( B ) ) ) )",
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( ( A B ) || ( foo? ( bar? ( ( C D || ( baz? ( E ) ( F G ) || ( H ) ) ) ) ) ) )",
				uselist=["foo", "bar", "baz"],
				expected_result=['||', [['A', 'B'], ['C', 'D', '||', ['E', ['F', 'G'], 'H']]]]),

			UseReduceTestCase(
				"|| ( ( A B ) || ( foo? ( bar? ( ( C D || ( baz? ( E ) ( F G ) || ( H ) ) ) ) ) ) )",
				uselist=["foo", "bar", "baz"],
				opconvert=True,
				expected_result=[['||', ['A', 'B'], ['C', 'D', ['||', 'E', ['F', 'G'], 'H']]]]),

			UseReduceTestCase(
				"|| ( foo? ( A B ) )",
				uselist=["foo"],
				opconvert=True,
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( || ( foo? ( A B ) ) )",
				uselist=["foo"],
				opconvert=True,
				expected_result=['A', 'B']),

			UseReduceTestCase(
				"|| ( || ( || ( a? ( b? ( c? ( || ( || ( || ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) )",
				uselist=["a", "b", "c", "d", "e", "f"],
				opconvert=True,
				expected_result=['A', 'B']),

			# flat test
			UseReduceTestCase(
				"A",
				flat=True,
				expected_result=["A"]),
			UseReduceTestCase(
				"( A )",
				flat=True,
				expected_result=["A"]),
			UseReduceTestCase(
				"|| ( A B )",
				flat=True,
				expected_result=["||", "A", "B"]),
			UseReduceTestCase(
				"|| ( A || ( B C ) )",
				flat=True,
				expected_result=["||", "A", "||", "B", "C"]),
			UseReduceTestCase(
				"|| ( A || ( B C D ) )",
				flat=True,
				expected_result=["||", "A", "||", "B", "C", "D"]),
			UseReduceTestCase(
				"|| ( A || ( B || ( C D ) E ) )",
				flat=True,
				expected_result=["||", "A", "||", "B", "||", "C", "D", "E"]),
			UseReduceTestCase(
				"( || ( ( ( A ) B ) ) )",
				flat=True,
				expected_result=["||", "A", "B"]),
			UseReduceTestCase(
				"( || ( || ( ( A ) B ) ) )",
				flat=True,
				expected_result=["||", "||", "A", "B"]),
			UseReduceTestCase(
				"( || ( || ( ( A ) B ) ) )",
				flat=True,
				expected_result=["||", "||", "A", "B"]),
			UseReduceTestCase(
				"|| ( A )",
				flat=True,
				expected_result=["||", "A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				expected_result=["A"]),
			UseReduceTestCase(
				"( || ( || ( || ( A ) foo? ( B ) ) ) )",
				uselist=["foo"],
				flat=True,
				expected_result=["||", "||", "||", "A", "B"]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				flat=True,
				expected_result=["||", "||", "||"]),
			UseReduceTestCase(
				"( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
				uselist=["foo", "bar"],
				flat=True,
				expected_result=["||", "||", "A", "||", "B"]),
			UseReduceTestCase(
				"A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B",
				flat=True,
				expected_result=["A", "||", "B"]),
			UseReduceTestCase(
				"|| ( A ) || ( B )",
				flat=True,
				expected_result=["||", "A", "||", "B"]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				flat=True,
				expected_result=[]),
			UseReduceTestCase(
				"foo? ( A ) foo? ( B )",
				uselist=["foo"],
				flat=True,
				expected_result=["A", "B"]),

			# use flag validation
			UseReduceTestCase(
				"foo? ( A )",
				uselist=["foo"],
				is_valid_flag=self.always_true,
				expected_result=["A"]),
			UseReduceTestCase(
				"foo? ( A )",
				is_valid_flag=self.always_true,
				expected_result=[]),

			# token_class
			UseReduceTestCase(
				"foo? ( dev-libs/A )",
				uselist=["foo"],
				token_class=Atom,
				expected_result=["dev-libs/A"]),
			UseReduceTestCase(
				"foo? ( dev-libs/A )",
				token_class=Atom,
				expected_result=[]),
		)

		test_cases_xfail = (
			UseReduceTestCase("? ( A )"),
			UseReduceTestCase("!? ( A )"),
			UseReduceTestCase("( A"),
			UseReduceTestCase("A )"),
			UseReduceTestCase("||( A B )"),
			UseReduceTestCase("|| (A B )"),
			UseReduceTestCase("|| ( A B)"),
			UseReduceTestCase("|| ( A B"),
			UseReduceTestCase("|| A B )"),
			UseReduceTestCase("|| A B"),
			UseReduceTestCase("|| ( A B ) )"),
			UseReduceTestCase("|| || B C"),
			UseReduceTestCase("|| ( A B || )"),
			UseReduceTestCase("a? A"),
			UseReduceTestCase("( || ( || || ( A ) foo? ( B ) ) )"),
			UseReduceTestCase("( || ( || bar? ( A ) foo? ( B ) ) )"),
			UseReduceTestCase("foo?"),
			UseReduceTestCase("foo? || ( A )"),
			UseReduceTestCase("|| ( )"),
			UseReduceTestCase("foo? ( )"),

			# SRC_URI stuff
			UseReduceTestCase("http://foo/bar -> blah.tbz2", is_src_uri=True, eapi=EAPI_WITHOUT_SRC_URI_ARROWS),
			UseReduceTestCase("|| ( http://foo/bar -> blah.tbz2 )", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar -> foo? ( ftp://foo/a )", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar blah.tbz2 ->", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("-> http://foo/bar blah.tbz2 )", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar ->", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar -> foo? ( http://foo.com/foo )", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("foo? ( http://foo/bar -> ) blah.tbz2", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar -> foo/blah.tbz2", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),
			UseReduceTestCase("http://foo/bar -> -> bar.tbz2 foo? ( ftp://foo/a )", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS),

			UseReduceTestCase("http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )", is_src_uri=False, eapi=EAPI_WITH_SRC_URI_ARROWS),

			UseReduceTestCase(
				"A",
				opconvert=True,
				flat=True),

			# use flag validation
			UseReduceTestCase("1.0? ( A )"),
			UseReduceTestCase("!1.0? ( A )"),
			UseReduceTestCase("!? ( A )"),
			UseReduceTestCase("!?? ( A )"),
			UseReduceTestCase(
				"foo? ( A )",
				is_valid_flag=self.always_false,
				),
			UseReduceTestCase(
				"foo? ( A )",
				uselist=["foo"],
				is_valid_flag=self.always_false,
				),

			# token_class
			UseReduceTestCase(
				"foo? ( A )",
				uselist=["foo"],
				token_class=Atom),
			UseReduceTestCase(
				"A(B",
				token_class=Atom),
		)

		for test_case in test_cases:
			# If it fails then show the input, since lots of our
			# test cases have the same output but different input,
			# making it difficult deduce which test has failed.
			self.assertEqual(test_case.run(), test_case.expected_result,
				"input: '%s' result: %s != %s" % (test_case.deparray,
				test_case.run(), test_case.expected_result))

		for test_case in test_cases_xfail:
			self.assertRaisesMsg(test_case.deparray, (InvalidDependString, ValueError), test_case.run)



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import get_required_use_flags
from portage.exception import InvalidDependString

class TestCheckRequiredUse(TestCase):

	def testCheckRequiredUse(self):
		test_cases = (
			("a b c", ["a", "b", "c"]),

			("|| ( a b c )", ["a", "b", "c"]),
			("^^ ( a b c )", ["a", "b", "c"]),
			("?? ( a b c )", ["a", "b", "c"]),
			("?? ( )", []),

			("|| ( a b ^^ ( d e f ) )", ["a", "b", "d", "e", "f"]),
			("^^ ( a b || ( d e f ) )", ["a", "b", "d", "e", "f"]),

			("( ^^ ( a ( b ) ( || ( ( d e ) ( f ) ) ) ) )", ["a", "b", "d", "e", "f"]),

			("a? ( ^^ ( b c ) )", ["a", "b", "c"]),
			("a? ( ^^ ( !b !d? ( c ) ) )", ["a", "b", "c", "d"]),
		)

		test_cases_xfail = (
			("^^ ( || ( a b ) ^^ ( b c )"),
			("^^( || ( a b ) ^^ ( b c ) )"),
			("^^ || ( a b ) ^^ ( b c )"),
			("^^ ( ( || ) ( a b ) ^^ ( b c ) )"),
			("^^ ( || ( a b ) ) ^^ ( b c ) )"),
		)

		for required_use, expected in test_cases:
			result = get_required_use_flags(required_use)
			expected = set(expected)
			self.assertEqual(result, expected, \
				"REQUIRED_USE: '%s', expected: '%s', got: '%s'" % (required_use, expected, result))

		for required_use in test_cases_xfail:
			self.assertRaisesMsg("REQUIRED_USE: '%s'" % (required_use,), \
				InvalidDependString, get_required_use_flags, required_use)



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (ResolverPlayground,
	ResolverPlaygroundTestCase)

class SonameUnsatisfiedTestCase(TestCase):

	def testSonameUnsatisfied(self):

		binpkgs = {
			"app-misc/A-1" : {
				"EAPI": "5",
				"PROVIDES": "x86_32: libA.so.1",
			},
			"app-misc/A-2" : {
				"EAPI": "5",
				"PROVIDES": "x86_32: libA.so.2",
			},
			"app-misc/B-0" : {
				"DEPEND": "app-misc/A",
				"RDEPEND": "app-misc/A",
				"REQUIRES": "x86_32: libA.so.2",
			}
		}

		installed = {
			"app-misc/A-2" : {
				"EAPI": "5",
				"PROVIDES": "x86_32: libA.so.2",
			},

			"app-misc/B-0" : {
				"DEPEND": "app-misc/A",
				"RDEPEND": "app-misc/A",
				"REQUIRES": "x86_32: libA.so.1",
			}
		}

		world = ["app-misc/B"]

		test_cases = (

			# Demonstrate bug #439694, where a broken
			# soname dependency needs to trigger a reinstall.
			ResolverPlaygroundTestCase(
				["@world"],
				options = {
					"--deep": True,
					"--ignore-soname-deps": "n",
					"--update": True,
					"--usepkgonly": True,
				},
				success = True,
				mergelist = [
					"[binary]app-misc/B-0"
				]
			),

			# This doesn't trigger a reinstall, since there's no version
			# change to trigger complete graph mode, and initially
			# unsatisfied deps are ignored in complete graph mode anyway.
			ResolverPlaygroundTestCase(
				["app-misc/A"],
				options = {
					"--ignore-soname-deps": "n",
					"--oneshot": True,
					"--usepkgonly": True,
				},
				success = True,
				mergelist = [
					"[binary]app-misc/A-2"
				]
			),
		)

		playground = ResolverPlayground(binpkgs=binpkgs, debug=False,
			installed=installed, world=world)
		try:
			for test_case in test_cases:
				playground.run_TestCase(test_case)
				self.assertEqual(test_case.test_success, True, test_case.fail_msg)
		finally:
			# Disable debug so that cleanup works.
			playground.debug = False
			playground.cleanup()



# Copyright 2014-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.repository.config import allow_profile_repo_deps
from portage.util import grabfile_package, stack_lists
from portage._sets.base import PackageSet

class ProfilePackageSet(PackageSet):
	_operations = ["merge"]

	def __init__(self, profiles, debug=False):
		super(ProfilePackageSet, self).__init__(
			allow_repo=any(allow_profile_repo_deps(y) for y in profiles)
		)
		self._profiles = profiles
		if profiles:
			desc_profile = profiles[-1]
			if desc_profile.user_config and len(profiles) > 1:
				desc_profile = profiles[-2]
			description = desc_profile.location
		else:
			description = None
		self.description = "Profile packages for profile %s" % description

	def load(self):
		self._setAtoms(x for x in stack_lists(
			[grabfile_package(os.path.join(y.location, "packages"),
			verify_eapi=True, eapi=y.eapi, eapi_default=None,
			allow_build_id=y.allow_build_id, allow_repo=allow_profile_repo_deps(y))
			for y in self._profiles
			if "profile-set" in y.profile_formats],
			incremental=1) if x[:1] != "*")

	def singleBuilder(self, options, settings, trees):
		return ProfilePackageSet(
			settings._locations_manager.profiles_complex)
	singleBuilder = classmethod(singleBuilder)



# Copyright 2005-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

doc = """Scan for failed merges and fix them."""
__doc__ = doc


module_spec = {
	'name': 'merges',
	'description': doc,
	'provides': {
		'merges': {
			'name': "merges",
			'sourcefile': "merges",
			'class': "MergesHandler",
			'description': doc,
			'functions': ['check', 'fix', 'purge'],
			'func_desc': {
				'purge': {
					'short': '-P', 'long': '--purge',
					'help': 'Removes the list of previously failed merges.' +
							' WARNING: Only use this option if you plan on' +
							' manually fixing them or do not want them'
							' re-installed.',
					'status': "Removing %s",
					'action': 'store_true',
					'func': 'purge'
				}
			},
			'opt_desc': {
				'yes': {
					"short": "-y",
					"long": "--yes",
					"help": ("(merges submodule only): Do not prompt for "
						"emerge invocations"),
					"action": "store_true",
					"dest": "yes",
				}
			},
		}
	}
}



# Copyright 1999-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

from _emerge.CompositeTask import CompositeTask
from portage import os
from portage.dbapi._MergeProcess import MergeProcess
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture

class EbuildMerge(CompositeTask):

	__slots__ = ("exit_hook", "find_blockers", "logger", "ldpath_mtimes",
		"pkg", "pkg_count", "pkg_path", "postinst_failure", "pretend",
		"settings", "tree", "world_atom")

	def _start(self):
		root_config = self.pkg.root_config
		settings = self.settings
		mycat = settings["CATEGORY"]
		mypkg = settings["PF"]
		pkgloc = settings["D"]
		infloc = os.path.join(settings["PORTAGE_BUILDDIR"], "build-info")
		myebuild = settings["EBUILD"]
		mydbapi = root_config.trees[self.tree].dbapi
		vartree = root_config.trees["vartree"]
		background = (settings.get('PORTAGE_BACKGROUND') == '1')
		logfile = settings.get('PORTAGE_LOG_FILE')

		merge_task = MergeProcess(
			mycat=mycat, mypkg=mypkg, settings=settings,
			treetype=self.tree, vartree=vartree, scheduler=self.scheduler,
			background=background, blockers=self.find_blockers, pkgloc=pkgloc,
			infloc=infloc, myebuild=myebuild, mydbapi=mydbapi,
			prev_mtimes=self.ldpath_mtimes, logfile=logfile)

		self._start_task(merge_task, self._merge_exit)

	def _merge_exit(self, merge_task):
		if self._final_exit(merge_task) != os.EX_OK:
			self._start_exit_hook(self.returncode)
			return

		self.postinst_failure = merge_task.postinst_failure
		pkg = self.pkg
		self.world_atom(pkg)
		pkg_count = self.pkg_count
		pkg_path = self.pkg_path
		logger = self.logger
		if "noclean" not in self.settings.features:
			short_msg = "emerge: (%s of %s) %s Clean Post" % \
				(pkg_count.curval, pkg_count.maxval, pkg.cpv)
			logger.log((" === (%s of %s) " + \
				"Post-Build Cleaning (%s::%s)") % \
				(pkg_count.curval, pkg_count.maxval, pkg.cpv, pkg_path),
				short_msg=short_msg)
		logger.log(" ::: completed emerge (%s of %s) %s to %s" % \
			(pkg_count.curval, pkg_count.maxval, pkg.cpv, pkg.root))

		self._start_exit_hook(self.returncode)

	def _start_exit_hook(self, returncode):
		"""
		Start the exit hook, and set returncode after it completes.
		"""
		# The returncode will be set after exit hook is complete.
		self.returncode = None
		self._start_task(
			AsyncTaskFuture(future=self.exit_hook(self)),
			functools.partial(self._exit_hook_exit, returncode))

	def _exit_hook_exit(self, returncode, task):
		self._assert_current(task)
		self.returncode = returncode
		self._async_wait()



# Copyright 2012-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import re
import textwrap

import portage
from portage import os
from portage.dep import Atom
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.update import update_dbentry
from portage.util import ensure_dirs
from portage.versions import _pkg_str
from portage._global_updates import _do_global_updates

class UpdateDbentryTestCase(TestCase):

	def testUpdateDbentryTestCase(self):
		cases = (

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "1",
				"  dev-libs/A:0  ", "  dev-libs/B:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "1",
				"  >=dev-libs/A-1:0  ", "  >=dev-libs/B-1:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "2",
				"  dev-libs/A[foo]  ", "  dev-libs/B[foo]  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "5",
				"  dev-libs/A:0/1=[foo]  ", "  dev-libs/B:0/1=[foo]  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "5",
				"  dev-libs/A:0/1[foo]  ", "  dev-libs/B:0/1[foo]  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "5",
				"  dev-libs/A:0/0[foo]  ", "  dev-libs/B:0/0[foo]  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")), "5",
				"  dev-libs/A:0=[foo]  ", "  dev-libs/B:0=[foo]  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "1",
				"  dev-libs/A:0  ", "  dev-libs/A:1  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "1",
				"  >=dev-libs/A-1:0  ", "  >=dev-libs/A-1:1  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "5",
				"  dev-libs/A:0/1=[foo]  ", "  dev-libs/A:1/1=[foo]  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "5",
				"  dev-libs/A:0/1[foo]  ", "  dev-libs/A:1/1[foo]  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "5",
				"  dev-libs/A:0/0[foo]  ", "  dev-libs/A:1/1[foo]  "),

			(("slotmove", Atom("dev-libs/A"), "0", "1"), "5",
				"  dev-libs/A:0=[foo]  ", "  dev-libs/A:1=[foo]  "),
		)
		for update_cmd, eapi, input_str, output_str in cases:
			result = update_dbentry(update_cmd, input_str, eapi=eapi)
			self.assertEqual(result, output_str)


	def testUpdateDbentryBlockerTestCase(self):
		"""
		Avoid creating self-blockers for bug #367215.
		"""
		cases = (

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/B-1", eapi="1", slot="0"),
				"  !dev-libs/A  ", "  !dev-libs/A  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/C-1", eapi="1", slot="0"),
				"  !dev-libs/A  ", "  !dev-libs/B  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/B-1", eapi="1", slot="0"),
				"  !dev-libs/A:0  ", "  !dev-libs/A:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/C-1", eapi="1", slot="0"),
				"  !dev-libs/A:0  ", "  !dev-libs/B:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/C-1", eapi="1", slot="0"),
				"  !>=dev-libs/A-1:0  ", "  !>=dev-libs/B-1:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/B-1", eapi="1", slot="0"),
				"  !>=dev-libs/A-1:0  ", "  !>=dev-libs/A-1:0  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/C-1", eapi="1", slot="0"),
				"  !>=dev-libs/A-1  ", "  !>=dev-libs/B-1  "),

			(("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
				_pkg_str("dev-libs/B-1", eapi="1", slot="0"),
				"  !>=dev-libs/A-1  ", "  !>=dev-libs/A-1  "),

		)
		for update_cmd, parent, input_str, output_str in cases:
			result = update_dbentry(update_cmd, input_str, parent=parent)
			self.assertEqual(result, output_str)

	def testUpdateDbentryDbapiTestCase(self):

		ebuilds = {

			"dev-libs/A-2::dont_apply_updates" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
				"SLOT": "2",
			},

			"dev-libs/B-2::dont_apply_updates" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
				"SLOT": "2",
			},

		}

		installed = {

			"dev-libs/A-1::test_repo" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
			},

			"dev-libs/A-2::dont_apply_updates" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
				"SLOT": "2",
			},

			"dev-libs/B-1::test_repo" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4-python",
			},

			"dev-libs/M-1::test_repo" : {
				"EAPI": "4",
			},

			"dev-libs/N-1::test_repo" : {
				"EAPI": "4",
			},

			"dev-libs/N-2::test_repo" : {
				"EAPI": "4-python",
			},

		}

		binpkgs = {

			"dev-libs/A-1::test_repo" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
			},

			"dev-libs/A-2::dont_apply_updates" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4",
				"SLOT": "2",
			},

			"dev-libs/B-1::test_repo" : {
				"RDEPEND" : "dev-libs/M dev-libs/N dev-libs/P",
				"EAPI": "4-python",
			},

		}

		world = ["dev-libs/M", "dev-libs/N"]

		updates = textwrap.dedent("""
			move dev-libs/M dev-libs/M-moved
			move dev-libs/N dev-libs/N.moved
		""")

		playground = ResolverPlayground(binpkgs=binpkgs,
			ebuilds=ebuilds, installed=installed, world=world)

		settings = playground.settings
		trees = playground.trees
		eroot = settings["EROOT"]
		test_repo_location = settings.repositories["test_repo"].location
		portdb = trees[eroot]["porttree"].dbapi
		vardb = trees[eroot]["vartree"].dbapi
		bindb = trees[eroot]["bintree"].dbapi
		setconfig = trees[eroot]["root_config"].setconfig
		selected_set = setconfig.getSets()["selected"]

		updates_dir = os.path.join(test_repo_location, "profiles", "updates")

		try:
			ensure_dirs(updates_dir)
			with open(os.path.join(updates_dir, "1Q-2010"), 'w') as f:
				f.write(updates)

			# Create an empty updates directory, so that this
			# repo doesn't inherit updates from the main repo.
			ensure_dirs(os.path.join(
				portdb.getRepositoryPath("dont_apply_updates"),
				"profiles", "updates"))

			global_noiselimit = portage.util.noiselimit
			portage.util.noiselimit = -2
			try:
				_do_global_updates(trees, {})
			finally:
				portage.util.noiselimit = global_noiselimit

			# Workaround for cache validation not working
			# correctly when filesystem has timestamp precision
			# of 1 second.
			vardb._clear_cache()

			# M -> M-moved
			old_pattern = re.compile(r"\bdev-libs/M(\s|$)")
			rdepend = vardb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/M-moved" in rdepend)
			rdepend = bindb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/M-moved" in rdepend)
			rdepend = vardb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/M-moved" in rdepend)
			rdepend = vardb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/M-moved" in rdepend)

			# EAPI 4-python/*-progress N -> N.moved
			rdepend = vardb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
			old_pattern = re.compile(r"\bdev-libs/N(\s|$)")
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/N.moved" in rdepend)
			rdepend = bindb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
			self.assertTrue(old_pattern.search(rdepend) is None)
			self.assertTrue("dev-libs/N.moved" in rdepend)
			self.assertRaises(KeyError,
				vardb.aux_get, "dev-libs/N-2", ["EAPI"])
			vardb.aux_get("dev-libs/N.moved-2", ["RDEPEND"])[0]

			# EAPI 4 does not allow dots in package names for N -> N.moved
			rdepend = vardb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
			self.assertTrue("dev-libs/N" in rdepend)
			self.assertTrue("dev-libs/N.moved" not in rdepend)
			rdepend = bindb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
			self.assertTrue("dev-libs/N" in rdepend)
			self.assertTrue("dev-libs/N.moved" not in rdepend)
			vardb.aux_get("dev-libs/N-1", ["RDEPEND"])[0]
			self.assertRaises(KeyError,
				vardb.aux_get, "dev-libs/N.moved-1", ["EAPI"])

			# dont_apply_updates
			rdepend = vardb.aux_get("dev-libs/A-2", ["RDEPEND"])[0]
			self.assertTrue("dev-libs/M" in rdepend)
			self.assertTrue("dev-libs/M-moved" not in rdepend)
			rdepend = bindb.aux_get("dev-libs/A-2", ["RDEPEND"])[0]
			self.assertTrue("dev-libs/M" in rdepend)
			self.assertTrue("dev-libs/M-moved" not in rdepend)

			selected_set.load()
			self.assertTrue("dev-libs/M" not in selected_set)
			self.assertTrue("dev-libs/M-moved" in selected_set)
			self.assertTrue("dev-libs/N" not in selected_set)
			self.assertTrue("dev-libs/N.moved" in selected_set)

		finally:
			playground.cleanup()



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools
import logging
import portage
from portage import os
from portage.dbapi._MergeProcess import MergeProcess
from portage.exception import UnsupportedAPIException
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from _emerge.EbuildBuildDir import EbuildBuildDir
from _emerge.emergelog import emergelog
from _emerge.CompositeTask import CompositeTask
from _emerge.unmerge import _unmerge_display

class PackageUninstall(CompositeTask):
	"""
	Uninstall a package asynchronously in a subprocess. When
	both parallel-install and ebuild-locks FEATURES are enabled,
	it is essential for the ebuild-locks code to execute in a
	subprocess, since the portage.locks module does not behave
	as desired if we try to lock the same file multiple times
	concurrently from the same process for ebuild-locks phases
	such as pkg_setup, pkg_prerm, and pkg_postrm.
	"""

	__slots__ = ("world_atom", "ldpath_mtimes", "opts",
			"pkg", "settings", "_builddir_lock")

	def _start(self):

		vardb = self.pkg.root_config.trees["vartree"].dbapi
		dbdir = vardb.getpath(self.pkg.cpv)
		if not os.path.exists(dbdir):
			# Apparently the package got uninstalled
			# already, so we can safely return early.
			self.returncode = os.EX_OK
			self._async_wait()
			return

		self.settings.setcpv(self.pkg)
		cat, pf = portage.catsplit(self.pkg.cpv)
		myebuildpath = os.path.join(dbdir, pf + ".ebuild")

		try:
			portage.doebuild_environment(myebuildpath, "prerm",
				settings=self.settings, db=vardb)
		except UnsupportedAPIException:
			# This is safe to ignore since this function is
			# guaranteed to set PORTAGE_BUILDDIR even though
			# it raises UnsupportedAPIException. The error
			# will be logged when it prevents the pkg_prerm
			# and pkg_postrm phases from executing.
			pass

		self._builddir_lock = EbuildBuildDir(
			scheduler=self.scheduler, settings=self.settings)
		self._start_task(
			AsyncTaskFuture(future=self._builddir_lock.async_lock()),
			self._start_unmerge)

	def _start_unmerge(self, lock_task):
		self._assert_current(lock_task)
		if lock_task.cancelled:
			self._default_final_exit(lock_task)
			return

		lock_task.future.result()
		portage.prepare_build_dirs(
			settings=self.settings, cleanup=True)

		# Output only gets logged if it comes after prepare_build_dirs()
		# which initializes PORTAGE_LOG_FILE.
		retval, _ = _unmerge_display(self.pkg.root_config,
			self.opts, "unmerge", [self.pkg.cpv], clean_delay=0,
			writemsg_level=self._writemsg_level)

		if retval != os.EX_OK:
			self._async_unlock_builddir(returncode=retval)
			return

		self._writemsg_level(">>> Unmerging %s...\n" % (self.pkg.cpv,),
			noiselevel=-1)
		self._emergelog("=== Unmerging... (%s)" % (self.pkg.cpv,))

		cat, pf = portage.catsplit(self.pkg.cpv)
		unmerge_task = MergeProcess(
			mycat=cat, mypkg=pf, settings=self.settings,
			treetype="vartree", vartree=self.pkg.root_config.trees["vartree"],
			scheduler=self.scheduler, background=self.background,
			mydbapi=self.pkg.root_config.trees["vartree"].dbapi,
			prev_mtimes=self.ldpath_mtimes,
			logfile=self.settings.get("PORTAGE_LOG_FILE"), unmerge=True)

		self._start_task(unmerge_task, self._unmerge_exit)

	def _unmerge_exit(self, unmerge_task):
		if self._final_exit(unmerge_task) != os.EX_OK:
			self._emergelog(" !!! unmerge FAILURE: %s" % (self.pkg.cpv,))
		else:
			self._emergelog(" >>> unmerge success: %s" % (self.pkg.cpv,))
			self.world_atom(self.pkg)
		self._async_unlock_builddir(returncode=self.returncode)

	def _async_unlock_builddir(self, returncode=None):
		"""
		Release the lock asynchronously, and if a returncode parameter
		is given then set self.returncode and notify exit listeners.
		"""
		if returncode is not None:
			# The returncode will be set after unlock is complete.
			self.returncode = None
		self._start_task(
			AsyncTaskFuture(future=self._builddir_lock.async_unlock()),
			functools.partial(self._unlock_builddir_exit, returncode=returncode))

	def _unlock_builddir_exit(self, unlock_task, returncode=None):
		self._assert_current(unlock_task)
		if unlock_task.cancelled and returncode is not None:
			self._default_final_exit(unlock_task)
			return

		# Normally, async_unlock should not raise an exception here.
		unlock_task.future.cancelled() or unlock_task.future.result()
		if returncode is not None:
			self.returncode = returncode
			self._async_wait()

	def _emergelog(self, msg):
		emergelog("notitles" not in self.settings.features, msg)

	def _writemsg_level(self, msg, level=0, noiselevel=0):

		log_path = self.settings.get("PORTAGE_LOG_FILE")
		background = self.background

		if log_path is None:
			if not (background and level < logging.WARNING):
				portage.util.writemsg_level(msg,
					level=level, noiselevel=noiselevel)
		else:
			self.scheduler.output(msg, log_path=log_path,
				level=level, noiselevel=noiselevel)



# getbinpkg.py -- Portage binary-package helper functions
# Copyright 2003-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.output import colorize
from portage.cache.mappings import slot_dict_class
from portage.localization import _
import portage
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.package.ebuild.fetch import _hide_url_passwd
from _emerge.Package import _all_metadata_keys

import pickle
import sys
import socket
import time
import tempfile
import base64
import warnings

_all_errors = [NotImplementedError, ValueError, socket.error]

from html.parser import HTMLParser as html_parser_HTMLParser
from urllib.parse import unquote as urllib_parse_unquote

try:
	import ftplib
except ImportError as e:
	sys.stderr.write(colorize("BAD", "!!! CANNOT IMPORT FTPLIB: ") + str(e) + "\n")
else:
	_all_errors.extend(ftplib.all_errors)

try:
	from http.client import HTTPConnection as http_client_HTTPConnection
	from http.client import BadStatusLine as http_client_BadStatusLine
	from http.client import ResponseNotReady as http_client_ResponseNotReady
	from http.client import error as http_client_error
except ImportError as e:
	sys.stderr.write(colorize("BAD", "!!! CANNOT IMPORT HTTP.CLIENT: ") + str(e) + "\n")
else:
	_all_errors.append(http_client_error)

_all_errors = tuple(_all_errors)


def make_metadata_dict(data):

	warnings.warn("portage.getbinpkg.make_metadata_dict() is deprecated",
		DeprecationWarning, stacklevel=2)

	myid, _myglob = data

	mydict = {}
	for k_bytes in portage.xpak.getindex_mem(myid):
		k = _unicode_decode(k_bytes,
			encoding=_encodings['repo.content'], errors='replace')
		if k not in _all_metadata_keys and k != "CATEGORY":
			continue
		v = _unicode_decode(portage.xpak.getitem(data, k_bytes),
			encoding=_encodings['repo.content'], errors='replace')
		mydict[k] = v

	return mydict

class ParseLinks(html_parser_HTMLParser):
	"""Parser class that overrides HTMLParser to grab all anchors from an html
	page and provide suffix and prefix limitors"""
	def __init__(self):

		warnings.warn("portage.getbinpkg.ParseLinks is deprecated",
			DeprecationWarning, stacklevel=2)

		self.PL_anchors = []
		html_parser_HTMLParser.__init__(self)

	def get_anchors(self):
		return self.PL_anchors

	def get_anchors_by_prefix(self, prefix):
		newlist = []
		for x in self.PL_anchors:
			if x.startswith(prefix):
				if x not in newlist:
					newlist.append(x[:])
		return newlist

	def get_anchors_by_suffix(self, suffix):
		newlist = []
		for x in self.PL_anchors:
			if x.endswith(suffix):
				if x not in newlist:
					newlist.append(x[:])
		return newlist

	def	handle_endtag(self, tag):
		pass

	def	handle_starttag(self, tag, attrs):
		if tag == "a":
			for x in attrs:
				if x[0] == 'href':
					if x[1] not in self.PL_anchors:
						self.PL_anchors.append(urllib_parse_unquote(x[1]))


def create_conn(baseurl, conn=None):
	"""Takes a protocol://site:port/address url, and an
	optional connection. If connection is already active, it is passed on.
	baseurl is reduced to address and is returned in tuple (conn,address)"""

	warnings.warn("portage.getbinpkg.create_conn() is deprecated",
		DeprecationWarning, stacklevel=2)

	parts = baseurl.split("://", 1)
	if len(parts) != 2:
		raise ValueError(_("Provided URI does not "
			"contain protocol identifier. '%s'") % baseurl)
	protocol, url_parts = parts
	del parts

	url_parts = url_parts.split("/")
	host = url_parts[0]
	if len(url_parts) < 2:
		address = "/"
	else:
		address = "/"+"/".join(url_parts[1:])
	del url_parts

	userpass_host = host.split("@", 1)
	if len(userpass_host) == 1:
		host = userpass_host[0]
		userpass = ["anonymous"]
	else:
		host = userpass_host[1]
		userpass = userpass_host[0].split(":")
	del userpass_host

	if len(userpass) > 2:
		raise ValueError(_("Unable to interpret username/password provided."))
	elif len(userpass) == 2:
		username = userpass[0]
		password = userpass[1]
	elif len(userpass) == 1:
		username = userpass[0]
		password = None
	del userpass

	http_headers = {}
	http_params = {}
	if username and password:
		try:
			encodebytes = base64.encodebytes
		except AttributeError:
			# Python 2
			encodebytes = base64.encodestring
		http_headers = {
			b"Authorization": "Basic %s" % \
			encodebytes(_unicode_encode("%s:%s" % (username, password))).replace(
			    b"\012",
			    b""
			  ),
		}

	if not conn:
		if protocol == "https":
			# Use local import since https typically isn't needed, and
			# this way we can usually avoid triggering the global scope
			# http.client ImportError handler (like during stage1 -> stage2
			# builds where USE=ssl is disabled for python).
			try:
				from http.client import HTTPSConnection as http_client_HTTPSConnection
			except ImportError:
				raise NotImplementedError(
					_("python must have ssl enabled for https support"))
			conn = http_client_HTTPSConnection(host)
		elif protocol == "http":
			conn = http_client_HTTPConnection(host)
		elif protocol == "ftp":
			passive = 1
			if host[-1] == "*":
				passive = 0
				host = host[:-1]
			conn = ftplib.FTP(host)
			if password:
				conn.login(username, password)
			else:
				sys.stderr.write(colorize("WARN",
					_(" * No password provided for username")) + " '%s'" % \
					(username,) + "\n\n")
				conn.login(username)
			conn.set_pasv(passive)
			conn.set_debuglevel(0)
		elif protocol == "sftp":
			try:
				import paramiko
			except ImportError:
				raise NotImplementedError(
					_("paramiko must be installed for sftp support"))
			t = paramiko.Transport(host)
			t.connect(username=username, password=password)
			conn = paramiko.SFTPClient.from_transport(t)
		else:
			raise NotImplementedError(_("%s is not a supported protocol.") % protocol)

	return (conn, protocol, address, http_params, http_headers)

def make_ftp_request(conn, address, rest=None, dest=None):
	"""Uses the |conn| object to request the data
	from address and issuing a rest if it is passed."""

	warnings.warn("portage.getbinpkg.make_ftp_request() is deprecated",
		DeprecationWarning, stacklevel=2)

	try:

		if dest:
			fstart_pos = dest.tell()

		conn.voidcmd("TYPE I")
		fsize = conn.size(address)

		if (rest != None) and (rest < 0):
			rest = fsize+int(rest)
		if rest < 0:
			rest = 0

		if rest != None:
			mysocket = conn.transfercmd("RETR %s" % str(address), rest)
		else:
			mysocket = conn.transfercmd("RETR %s" % str(address))

		mydata = ""
		while 1:
			somedata = mysocket.recv(8192)
			if somedata:
				if dest:
					dest.write(somedata)
				else:
					mydata = mydata + somedata
			else:
				break

		if dest:
			data_size = fstart_pos - dest.tell()
		else:
			data_size = len(mydata)

		mysocket.close()
		conn.voidresp()
		conn.voidcmd("TYPE A")

		return mydata, (fsize != data_size), ""

	except ValueError as e:
		return None, int(str(e)[:4]), str(e)


def make_http_request(conn, address, _params={}, headers={}, dest=None):
	"""Uses the |conn| object to request
	the data from address, performing Location forwarding and using the
	optional params and headers."""

	warnings.warn("portage.getbinpkg.make_http_request() is deprecated",
		DeprecationWarning, stacklevel=2)

	rc = 0
	response = None
	while (rc == 0) or (rc == 301) or (rc == 302):
		try:
			if rc != 0:
				conn = create_conn(address)[0]
			conn.request("GET", address, body=None, headers=headers)
		except SystemExit as e:
			raise
		except Exception as e:
			return None, None, "Server request failed: %s" % str(e)
		response = conn.getresponse()
		rc = response.status

		# 301 means that the page address is wrong.
		if ((rc == 301) or (rc == 302)):
			ignored_data = response.read()
			del ignored_data
			for x in str(response.msg).split("\n"):
				parts = x.split(": ", 1)
				if parts[0] == "Location":
					if rc == 301:
						sys.stderr.write(colorize("BAD",
							_("Location has moved: ")) + str(parts[1]) + "\n")
					if rc == 302:
						sys.stderr.write(colorize("BAD",
							_("Location has temporarily moved: ")) + \
							str(parts[1]) + "\n")
					address = parts[1]
					break

	if (rc != 200) and (rc != 206):
		return None, rc, "Server did not respond successfully (%s: %s)" % (str(response.status), str(response.reason))

	if dest:
		dest.write(response.read())
		return "", 0, ""

	return response.read(), 0, ""


def match_in_array(array, prefix="", suffix="", match_both=1, allow_overlap=0):

	warnings.warn("portage.getbinpkg.match_in_array() is deprecated",
		DeprecationWarning, stacklevel=2)

	myarray = []

	if not (prefix and suffix):
		match_both = 0

	for x in array:
		add_p = 0
		if prefix and (len(x) >= len(prefix)) and (x[:len(prefix)] == prefix):
			add_p = 1

		if match_both:
			if prefix and not add_p: # Require both, but don't have first one.
				continue
		else:
			if add_p:     # Only need one, and we have it.
				myarray.append(x[:])
				continue

		if not allow_overlap: # Not allow to overlap prefix and suffix
			if len(x) >= (len(prefix)+len(suffix)):
				pass
			else:
				continue          # Too short to match.
		else:
			pass                      # Do whatever... We're overlapping.

		if suffix and (len(x) >= len(suffix)) and (x[-len(suffix):] == suffix):
			myarray.append(x)   # It matches
		else:
			continue            # Doesn't match.

	return myarray


def dir_get_list(baseurl, conn=None):
	"""Takes a base url to connect to and read from.
	URI should be in the form <proto>://<site>[:port]<path>
	Connection is used for persistent connection instances."""

	warnings.warn("portage.getbinpkg.dir_get_list() is deprecated",
		DeprecationWarning, stacklevel=2)

	if not conn:
		keepconnection = 0
	else:
		keepconnection = 1

	conn, protocol, address, params, headers = create_conn(baseurl, conn)

	listing = None
	if protocol in ["http","https"]:
		if not address.endswith("/"):
			# http servers can return a 400 error here
			# if the address doesn't end with a slash.
			address += "/"
		page, rc, msg = make_http_request(conn, address, params, headers)

		if page:
			parser = ParseLinks()
			parser.feed(_unicode_decode(page))
			del page
			listing = parser.get_anchors()
		else:
			import portage.exception
			raise portage.exception.PortageException(
				_("Unable to get listing: %s %s") % (rc,msg))
	elif protocol in ["ftp"]:
		if address[-1] == '/':
			olddir = conn.pwd()
			conn.cwd(address)
			listing = conn.nlst()
			conn.cwd(olddir)
			del olddir
		else:
			listing = conn.nlst(address)
	elif protocol == "sftp":
		listing = conn.listdir(address)
	else:
		raise TypeError(_("Unknown protocol. '%s'") % protocol)

	if not keepconnection:
		conn.close()

	return listing

def file_get_metadata(baseurl, conn=None, chunk_size=3000):
	"""Takes a base url to connect to and read from.
	URI should be in the form <proto>://<site>[:port]<path>
	Connection is used for persistent connection instances."""

	warnings.warn("portage.getbinpkg.file_get_metadata() is deprecated",
		DeprecationWarning, stacklevel=2)

	if not conn:
		keepconnection = 0
	else:
		keepconnection = 1

	conn, protocol, address, params, headers = create_conn(baseurl, conn)

	if protocol in ["http","https"]:
		headers["Range"] = "bytes=-%s" % str(chunk_size)
		data, _x, _x = make_http_request(conn, address, params, headers)
	elif protocol in ["ftp"]:
		data, _x, _x = make_ftp_request(conn, address, -chunk_size)
	elif protocol == "sftp":
		f = conn.open(address)
		try:
			f.seek(-chunk_size, 2)
			data = f.read()
		finally:
			f.close()
	else:
		raise TypeError(_("Unknown protocol. '%s'") % protocol)

	if data:
		xpaksize = portage.xpak.decodeint(data[-8:-4])
		if (xpaksize + 8) > chunk_size:
			myid = file_get_metadata(baseurl, conn, xpaksize + 8)
			if not keepconnection:
				conn.close()
			return myid
		xpak_data = data[len(data) - (xpaksize + 8):-8]
		del data

		myid = portage.xpak.xsplit_mem(xpak_data)
		if not myid:
			myid = None, None
		del xpak_data
	else:
		myid = None, None

	if not keepconnection:
		conn.close()

	return myid


def file_get(baseurl=None, dest=None, conn=None, fcmd=None, filename=None,
	fcmd_vars=None):
	"""Takes a base url to connect to and read from.
	URI should be in the form <proto>://[user[:pass]@]<site>[:port]<path>"""

	if not fcmd:

		warnings.warn("Use of portage.getbinpkg.file_get() without the fcmd "
			"parameter is deprecated", DeprecationWarning, stacklevel=2)

		return file_get_lib(baseurl, dest, conn)

	variables = {}

	if fcmd_vars is not None:
		variables.update(fcmd_vars)

	if "DISTDIR" not in variables:
		if dest is None:
			raise portage.exception.MissingParameter(
				_("%s is missing required '%s' key") %
				("fcmd_vars", "DISTDIR"))
		variables["DISTDIR"] = dest

	if "URI" not in variables:
		if baseurl is None:
			raise portage.exception.MissingParameter(
				_("%s is missing required '%s' key") %
				("fcmd_vars", "URI"))
		variables["URI"] = baseurl

	if "FILE" not in variables:
		if filename is None:
			filename = os.path.basename(variables["URI"])
		variables["FILE"] = filename

	from portage.util import varexpand
	from portage.process import spawn
	myfetch = portage.util.shlex_split(fcmd)
	myfetch = [varexpand(x, mydict=variables) for x in myfetch]
	fd_pipes = {
		0: portage._get_stdin().fileno(),
		1: sys.__stdout__.fileno(),
		2: sys.__stdout__.fileno()
	}
	sys.__stdout__.flush()
	sys.__stderr__.flush()
	retval = spawn(myfetch, env=os.environ.copy(), fd_pipes=fd_pipes)
	if retval != os.EX_OK:
		sys.stderr.write(_("Fetcher exited with a failure condition.\n"))
		return 0
	return 1

def file_get_lib(baseurl, dest, conn=None):
	"""Takes a base url to connect to and read from.
	URI should be in the form <proto>://<site>[:port]<path>
	Connection is used for persistent connection instances."""

	warnings.warn("portage.getbinpkg.file_get_lib() is deprecated",
		DeprecationWarning, stacklevel=2)

	if not conn:
		keepconnection = 0
	else:
		keepconnection = 1

	conn, protocol, address, params, headers = create_conn(baseurl, conn)

	sys.stderr.write("Fetching '" + str(os.path.basename(address)) + "'\n")
	if protocol in ["http", "https"]:
		data, rc, _msg = make_http_request(conn, address, params, headers, dest=dest)
	elif protocol in ["ftp"]:
		data, rc, _msg = make_ftp_request(conn, address, dest=dest)
	elif protocol == "sftp":
		rc = 0
		try:
			f = conn.open(address)
		except SystemExit:
			raise
		except Exception:
			rc = 1
		else:
			try:
				if dest:
					bufsize = 8192
					while True:
						data = f.read(bufsize)
						if not data:
							break
						dest.write(data)
			finally:
				f.close()
	else:
		raise TypeError(_("Unknown protocol. '%s'") % protocol)

	if not keepconnection:
		conn.close()

	return rc


def dir_get_metadata(baseurl, conn=None, chunk_size=3000, verbose=1, usingcache=1, makepickle=None):

	warnings.warn("portage.getbinpkg.dir_get_metadata() is deprecated",
		DeprecationWarning, stacklevel=2)

	if not conn:
		keepconnection = 0
	else:
		keepconnection = 1

	cache_path = "/var/cache/edb"
	metadatafilename = os.path.join(cache_path, 'remote_metadata.pickle')

	if makepickle is None:
		makepickle = "/var/cache/edb/metadata.idx.most_recent"

	try:
		conn = create_conn(baseurl, conn)[0]
	except _all_errors as e:
		# ftplib.FTP(host) can raise errors like this:
		#   socket.error: (111, 'Connection refused')
		sys.stderr.write("!!! %s\n" % (e,))
		return {}

	out = sys.stdout
	try:
		metadatafile = open(_unicode_encode(metadatafilename,
			encoding=_encodings['fs'], errors='strict'), 'rb')
		mypickle = pickle.Unpickler(metadatafile)
		try:
			mypickle.find_global = None
		except AttributeError:
			# TODO: If py3k, override Unpickler.find_class().
			pass
		metadata = mypickle.load()
		out.write(_("Loaded metadata pickle.\n"))
		out.flush()
		metadatafile.close()
	except (SystemExit, KeyboardInterrupt):
		raise
	except Exception:
		metadata = {}
	if baseurl not in metadata:
		metadata[baseurl] = {}
	if "indexname" not in metadata[baseurl]:
		metadata[baseurl]["indexname"] = ""
	if "timestamp" not in metadata[baseurl]:
		metadata[baseurl]["timestamp"] = 0
	if "unmodified" not in metadata[baseurl]:
		metadata[baseurl]["unmodified"] = 0
	if "data" not in metadata[baseurl]:
		metadata[baseurl]["data"] = {}

	if not os.access(cache_path, os.W_OK):
		sys.stderr.write(_("!!! Unable to write binary metadata to disk!\n"))
		sys.stderr.write(_("!!! Permission denied: '%s'\n") % cache_path)
		return metadata[baseurl]["data"]

	import portage.exception
	try:
		filelist = dir_get_list(baseurl, conn)
	except portage.exception.PortageException as e:
		sys.stderr.write(_("!!! Error connecting to '%s'.\n") %
			_hide_url_passwd(baseurl))
		sys.stderr.write("!!! %s\n" % str(e))
		del e
		return metadata[baseurl]["data"]
	tbz2list = match_in_array(filelist, suffix=".tbz2")
	metalist = match_in_array(filelist, prefix="metadata.idx")
	del filelist

	# Determine if our metadata file is current.
	metalist.sort()
	metalist.reverse() # makes the order new-to-old.
	for mfile in metalist:
		if usingcache and \
		   ((metadata[baseurl]["indexname"] != mfile) or \
			  (metadata[baseurl]["timestamp"] < int(time.time() - (60 * 60 * 24)))):
			# Try to download new cache until we succeed on one.
			data = ""
			for trynum in [1, 2, 3]:
				mytempfile = tempfile.TemporaryFile()
				try:
					file_get(baseurl + "/" + mfile, mytempfile, conn)
					if mytempfile.tell() > len(data):
						mytempfile.seek(0)
						data = mytempfile.read()
				except ValueError as e:
					sys.stderr.write("--- %s\n" % str(e))
					if trynum < 3:
						sys.stderr.write(_("Retrying...\n"))
					sys.stderr.flush()
					mytempfile.close()
					continue
				if match_in_array([mfile], suffix=".gz"):
					out.write("gzip'd\n")
					out.flush()
					try:
						import gzip
						mytempfile.seek(0)
						gzindex = gzip.GzipFile(mfile[:-3], 'rb', 9, mytempfile)
						data = gzindex.read()
					except SystemExit as e:
						raise
					except Exception as e:
						mytempfile.close()
						sys.stderr.write(_("!!! Failed to use gzip: ") + str(e) + "\n")
						sys.stderr.flush()
					mytempfile.close()
				try:
					metadata[baseurl]["data"] = pickle.loads(data)
					del data
					metadata[baseurl]["indexname"] = mfile
					metadata[baseurl]["timestamp"] = int(time.time())
					metadata[baseurl]["modified"]  = 0 # It's not, right after download.
					out.write(_("Pickle loaded.\n"))
					out.flush()
					break
				except SystemExit as e:
					raise
				except Exception as e:
					sys.stderr.write(_("!!! Failed to read data from index: ") + str(mfile) + "\n")
					sys.stderr.write("!!! %s" % str(e))
					sys.stderr.flush()
			try:
				metadatafile = open(_unicode_encode(metadatafilename,
					encoding=_encodings['fs'], errors='strict'), 'wb')
				pickle.dump(metadata, metadatafile, protocol=2)
				metadatafile.close()
			except SystemExit as e:
				raise
			except Exception as e:
				sys.stderr.write(_("!!! Failed to write binary metadata to disk!\n"))
				sys.stderr.write("!!! %s\n" % str(e))
				sys.stderr.flush()
			break
	# We may have metadata... now we run through the tbz2 list and check.

	class CacheStats:
		from time import time
		def __init__(self, out):
			self.misses = 0
			self.hits = 0
			self.last_update = 0
			self.out = out
			self.min_display_latency = 0.2
		def update(self):
			cur_time = self.time()
			if cur_time - self.last_update >= self.min_display_latency:
				self.last_update = cur_time
				self.display()
		def display(self):
			self.out.write("\r"+colorize("WARN",
				_("cache miss: '") + str(self.misses) + "'") + \
				" --- " + colorize("GOOD", _("cache hit: '") + str(self.hits) + "'"))
			self.out.flush()

	cache_stats = CacheStats(out)
	have_tty = os.environ.get('TERM') != 'dumb' and out.isatty()
	if have_tty:
		cache_stats.display()
	binpkg_filenames = set()
	for x in tbz2list:
		x = os.path.basename(x)
		binpkg_filenames.add(x)
		if x not in metadata[baseurl]["data"]:
			cache_stats.misses += 1
			if have_tty:
				cache_stats.update()
			metadata[baseurl]["modified"] = 1
			myid = None
			for _x in range(3):
				try:
					myid = file_get_metadata(
						"/".join((baseurl.rstrip("/"), x.lstrip("/"))),
						conn, chunk_size)
					break
				except http_client_BadStatusLine:
					# Sometimes this error is thrown from conn.getresponse() in
					# make_http_request().  The docstring for this error in
					# httplib.py says "Presumably, the server closed the
					# connection before sending a valid response".
					conn = create_conn(baseurl)[0]
				except http_client_ResponseNotReady:
					# With some http servers this error is known to be thrown
					# from conn.getresponse() in make_http_request() when the
					# remote file does not have appropriate read permissions.
					# Maybe it's possible to recover from this exception in
					# cases though, so retry.
					conn = create_conn(baseurl)[0]

			if myid and myid[0]:
				metadata[baseurl]["data"][x] = make_metadata_dict(myid)
			elif verbose:
				sys.stderr.write(colorize("BAD",
					_("!!! Failed to retrieve metadata on: ")) + str(x) + "\n")
				sys.stderr.flush()
		else:
			cache_stats.hits += 1
			if have_tty:
				cache_stats.update()
	cache_stats.display()
	# Cleanse stale cache for files that don't exist on the server anymore.
	stale_cache = set(metadata[baseurl]["data"]).difference(binpkg_filenames)
	if stale_cache:
		for x in stale_cache:
			del metadata[baseurl]["data"][x]
		metadata[baseurl]["modified"] = 1
	del stale_cache
	del binpkg_filenames
	out.write("\n")
	out.flush()

	try:
		if "modified" in metadata[baseurl] and metadata[baseurl]["modified"]:
			metadata[baseurl]["timestamp"] = int(time.time())
			metadatafile = open(_unicode_encode(metadatafilename,
				encoding=_encodings['fs'], errors='strict'), 'wb')
			pickle.dump(metadata, metadatafile, protocol=2)
			metadatafile.close()
		if makepickle:
			metadatafile = open(_unicode_encode(makepickle,
				encoding=_encodings['fs'], errors='strict'), 'wb')
			pickle.dump(metadata[baseurl]["data"], metadatafile, protocol=2)
			metadatafile.close()
	except SystemExit as e:
		raise
	except Exception as e:
		sys.stderr.write(_("!!! Failed to write binary metadata to disk!\n"))
		sys.stderr.write("!!! "+str(e)+"\n")
		sys.stderr.flush()

	if not keepconnection:
		conn.close()

	return metadata[baseurl]["data"]

def _cmp_cpv(d1, d2):
	cpv1 = d1["CPV"]
	cpv2 = d2["CPV"]
	if cpv1 > cpv2:
		return 1
	if cpv1 == cpv2:
		return 0
	return -1

class PackageIndex:

	def __init__(self,
		allowed_pkg_keys=None,
		default_header_data=None,
		default_pkg_data=None,
		inherited_keys=None,
		translated_keys=None):

		self._pkg_slot_dict = None
		if allowed_pkg_keys is not None:
			self._pkg_slot_dict = slot_dict_class(allowed_pkg_keys)

		self._default_header_data = default_header_data
		self._default_pkg_data = default_pkg_data
		self._inherited_keys = inherited_keys
		self._write_translation_map = {}
		self._read_translation_map = {}
		if translated_keys:
			self._write_translation_map.update(translated_keys)
			self._read_translation_map.update(((y, x) for (x, y) in translated_keys))
		self.header = {}
		if self._default_header_data:
			self.header.update(self._default_header_data)
		self.packages = []
		self.modified = True

	def _readpkgindex(self, pkgfile, pkg_entry=True):

		allowed_keys = None
		if self._pkg_slot_dict is None or not pkg_entry:
			d = {}
		else:
			d = self._pkg_slot_dict()
			allowed_keys = d.allowed_keys

		for line in pkgfile:
			line = line.rstrip("\n")
			if not line:
				break
			line = line.split(":", 1)
			if not len(line) == 2:
				continue
			k, v = line
			if v:
				v = v[1:]
			k = self._read_translation_map.get(k, k)
			if allowed_keys is not None and \
				k not in allowed_keys:
				continue
			d[k] = v
		return d

	def _writepkgindex(self, pkgfile, items):
		for k, v in items:
			pkgfile.write("%s: %s\n" % \
				(self._write_translation_map.get(k, k), v))
		pkgfile.write("\n")

	def read(self, pkgfile):
		self.readHeader(pkgfile)
		self.readBody(pkgfile)

	def readHeader(self, pkgfile):
		self.header.update(self._readpkgindex(pkgfile, pkg_entry=False))

	def readBody(self, pkgfile):
		while True:
			d = self._readpkgindex(pkgfile)
			if not d:
				break
			mycpv = d.get("CPV")
			if not mycpv:
				continue
			if self._default_pkg_data:
				for k, v in self._default_pkg_data.items():
					d.setdefault(k, v)
			if self._inherited_keys:
				for k in self._inherited_keys:
					v = self.header.get(k)
					if v is not None:
						d.setdefault(k, v)
			self.packages.append(d)

	def write(self, pkgfile):
		if self.modified:
			self.header["TIMESTAMP"] = str(int(time.time()))
			self.header["PACKAGES"] = str(len(self.packages))
		keys = list(self.header)
		keys.sort()
		self._writepkgindex(pkgfile, [(k, self.header[k]) \
			for k in keys if self.header[k]])
		for metadata in sorted(self.packages,
			key=portage.util.cmp_sort_key(_cmp_cpv)):
			metadata = metadata.copy()
			if self._inherited_keys:
				for k in self._inherited_keys:
					v = self.header.get(k)
					if v is not None and v == metadata.get(k):
						del metadata[k]
			if self._default_pkg_data:
				for k, v in self._default_pkg_data.items():
					if metadata.get(k) == v:
						metadata.pop(k, None)
			keys = list(metadata)
			keys.sort()
			self._writepkgindex(pkgfile,
				[(k, metadata[k]) for k in keys if metadata[k]])



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.output import bold, turquoise, green

def emerge_help():
	print(bold("emerge:")+" command-line interface to the Portage system")
	print(bold("Usage:"))
	print("   "+turquoise("emerge")+" [ "+green("options")+" ] [ "+green("action")+" ] [ "+turquoise("ebuild")+" | "+turquoise("tbz2")+" | "+turquoise("file")+" | "+turquoise("@set")+" | "+turquoise("atom")+" ] [ ... ]")
	print("   "+turquoise("emerge")+" [ "+green("options")+" ] [ "+green("action")+" ] < "+turquoise("@system")+" | "+turquoise("@world")+" >")
	print("   "+turquoise("emerge")+" < "+turquoise("--sync")+" | "+turquoise("--metadata")+" | "+turquoise("--info")+" >")
	print("   "+turquoise("emerge")+" "+turquoise("--resume")+" [ "+green("--pretend")+" | "+green("--ask")+" | "+green("--skipfirst")+" ]")
	print("   "+turquoise("emerge")+" "+turquoise("--help"))
	print(bold("Options:")+" "+green("-")+"["+green("abBcCdDefgGhjkKlnNoOpPqrsStuUvVwW")+"]")
	print("          [ " + green("--color")+" < " + turquoise("y") + " | "+ turquoise("n")+" >            ] [ "+green("--columns")+"    ]")
	print("          [ "+green("--complete-graph")+"             ] [ "+green("--deep")+"       ]")
	print("          [ "+green("--jobs") + " " + turquoise("JOBS")+" ] [ "+green("--keep-going")+" ] [ " + green("--load-average")+" " + turquoise("LOAD") + "            ]")
	print("          [ "+green("--newrepo")+"   ] [ "+green("--newuse")+"     ] [ "+green("--noconfmem")+"  ] [ "+green("--nospinner")+"   ]")
	print("          [ "+green("--oneshot")+"   ] [ "+green("--onlydeps")+"   ] [ "+ green("--quiet-build")+" [ " + turquoise("y") + " | "+ turquoise("n")+" ]        ]")
	print("          [ "+green("--reinstall ")+turquoise("changed-use")+"      ] [ " + green("--with-bdeps")+" < " + turquoise("y") + " | "+ turquoise("n")+" >         ]")
	print(bold("Actions:")+"  [ "+green("--depclean")+" | "+green("--list-sets")+" | "+green("--search")+" | "+green("--sync")+" | "+green("--version")+"        ]")
	print()
	print("   For more help consult the man page.")



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ['deprecated_profile_check']

import io

import portage
from portage import os, _encodings, _unicode_encode
from portage.const import DEPRECATED_PROFILE_FILE
from portage.localization import _
from portage.output import colorize
from portage.util import writemsg

def deprecated_profile_check(settings=None):
	config_root = None
	eprefix = None
	deprecated_profile_file = None
	if settings is not None:
		config_root = settings["PORTAGE_CONFIGROOT"]
		eprefix = settings["EPREFIX"]
		for x in reversed(settings._locations_manager.profiles_complex):
			if x.show_deprecated_warning:
				deprecated_profile_file = os.path.join(x.location, "deprecated")
				if os.access(deprecated_profile_file, os.R_OK):
					break
		else:
			deprecated_profile_file = None

	if deprecated_profile_file is None:
		deprecated_profile_file = os.path.join(config_root or "/",
			DEPRECATED_PROFILE_FILE)
		if not os.access(deprecated_profile_file, os.R_OK):
			deprecated_profile_file = os.path.join(config_root or "/",
				'etc', 'make.profile', 'deprecated')
			if not os.access(deprecated_profile_file, os.R_OK):
				return

	with io.open(_unicode_encode(deprecated_profile_file,
		encoding=_encodings['fs'], errors='strict'),
		mode='r', encoding=_encodings['content'], errors='replace') as f:
		dcontent = f.readlines()
	writemsg(colorize("BAD", _("\n!!! Your current profile is "
		"deprecated and not supported anymore.")) + "\n", noiselevel=-1)
	writemsg(colorize("BAD", _("!!! Use eselect profile to update your "
		"profile.")) + "\n", noiselevel=-1)
	if not dcontent:
		writemsg(colorize("BAD", _("!!! Please refer to the "
			"Gentoo Upgrading Guide.")) + "\n", noiselevel=-1)
		return True
	newprofile = dcontent[0].rstrip("\n")
	writemsg(colorize("BAD", _("!!! Please upgrade to the "
		"following profile if possible:")) + "\n\n", noiselevel=-1)
	writemsg(8*" " + colorize("GOOD", newprofile) + "\n\n", noiselevel=-1)
	if len(dcontent) > 1:
		writemsg(_("To upgrade do the following steps:\n"), noiselevel=-1)
		for myline in dcontent[1:]:
			writemsg(myline, noiselevel=-1)
		writemsg("\n\n", noiselevel=-1)
	else:
		writemsg(_("You may use the following command to upgrade:\n\n"), noiselevel=-1)
		writemsg(8*" " + colorize("INFORM", 'eselect profile set ' +
			newprofile) + "\n\n", noiselevel=-1)

	if settings is not None:
		main_repo_loc = settings.repositories.mainRepoLocation()
		new_profile_path = os.path.join(main_repo_loc,
			"profiles", newprofile.rstrip("\n"))

		if os.path.isdir(new_profile_path):
			new_config = portage.config(config_root=config_root,
				config_profile_path=new_profile_path,
				eprefix=eprefix)

			if not new_config.profiles:
				writemsg("\n %s %s\n" % (colorize("WARN", "*"),
					_("You must update portage before you "
					"can migrate to the above profile.")), noiselevel=-1)
				writemsg(" %s %s\n\n" % (colorize("WARN", "*"),
					_("In order to update portage, "
					"run 'emerge --oneshot sys-apps/portage'.")),
					noiselevel=-1)

	return True



# Copyright 2010-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractEbuildProcess import AbstractEbuildProcess

class EbuildSpawnProcess(AbstractEbuildProcess):
	"""
	Used by doebuild.spawn() to manage the spawned process.
	"""
	_spawn_kwarg_names = AbstractEbuildProcess._spawn_kwarg_names + \
		('fakeroot_state',)

	__slots__ = ('fakeroot_state', 'spawn_func')

	def _spawn(self, args, **kwargs):

		env = self.settings.environ()

		if self._dummy_pipe_fd is not None:
			env["PORTAGE_PIPE_FD"] = str(self._dummy_pipe_fd)

		return self.spawn_func(args, env=env, **kwargs)



# Copyright 2011-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile
import traceback

import portage
from portage import os
from portage import shutil
from portage.exception import TryAgain
from portage.tests import TestCase

class LockNonblockTestCase(TestCase):

	def _testLockNonblock(self):
		tempdir = tempfile.mkdtemp()
		try:
			path = os.path.join(tempdir, 'lock_me')
			lock1 = portage.locks.lockfile(path)
			pid = os.fork()
			if pid == 0:
				portage._ForkWatcher.hook(portage._ForkWatcher)
				portage.locks._close_fds()
				 # Disable close_fds since we don't exec
				 # (see _setup_pipes docstring).
				portage.process._setup_pipes({0:0, 1:1, 2:2}, close_fds=False)
				rval = 2
				try:
					try:
						lock2 = portage.locks.lockfile(path, flags=os.O_NONBLOCK)
					except portage.exception.TryAgain:
						rval = os.EX_OK
					else:
						rval = 1
						portage.locks.unlockfile(lock2)
				except SystemExit:
					raise
				except:
					traceback.print_exc()
				finally:
					os._exit(rval)

			self.assertEqual(pid > 0, True)
			pid, status = os.waitpid(pid, 0)
			self.assertEqual(os.WIFEXITED(status), True)
			self.assertEqual(os.WEXITSTATUS(status), os.EX_OK)

			portage.locks.unlockfile(lock1)
		finally:
			shutil.rmtree(tempdir)

	def testLockNonblock(self):
		self._testLockNonblock()

	def testLockNonblockHardlink(self):
		prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
		os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
		try:
			self._testLockNonblock()
		finally:
			os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
			if prev_state is not None:
				os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

	def test_competition_with_same_process(self):
		"""
		Test that at attempt to lock the same file multiple times in the
		same process will behave as intended (bug 714480).
		"""
		tempdir = tempfile.mkdtemp()
		try:
			path = os.path.join(tempdir, 'lock_me')
			lock = portage.locks.lockfile(path)
			self.assertRaises(TryAgain, portage.locks.lockfile, path, flags=os.O_NONBLOCK)
			portage.locks.unlockfile(lock)
		finally:
			shutil.rmtree(tempdir)



# Copyright 2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (ResolverPlayground,
	ResolverPlaygroundTestCase)

class RebuildTestCase(TestCase):

	def testRebuild(self):
		"""
		Rebuild packages when build-time dependencies are upgraded.
		"""

		ebuilds = {
			"sys-libs/x-1": { },
			"sys-libs/x-1-r1": { },
			"sys-libs/x-2": { },
			"sys-apps/a-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/a-2": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/b-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/b-2": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/c-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/c-2": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/d-1": { "RDEPEND" : "sys-libs/x"},
			"sys-apps/d-2": { "RDEPEND" : "sys-libs/x"},
			"sys-apps/e-2": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/f-2": { "DEPEND"  : "sys-apps/a", "RDEPEND" : ""},
			"sys-apps/g-2": { "DEPEND"  : "sys-apps/b sys-libs/x",
				"RDEPEND" : ""},
			}

		installed = {
			"sys-libs/x-1": { },
			"sys-apps/a-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/b-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/c-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/d-1": { "RDEPEND" : "sys-libs/x"},
			"sys-apps/e-1": { "DEPEND"  : "sys-libs/x", "RDEPEND" : ""},
			"sys-apps/f-1": { "DEPEND"  : "sys-apps/a", "RDEPEND" : ""},
			"sys-apps/g-1": { "DEPEND"  : "sys-apps/b",
				"RDEPEND" : ""},
			}

		world = ["sys-apps/a", "sys-apps/b", "sys-apps/c", "sys-apps/d",
			"sys-apps/e", "sys-apps/f", "sys-apps/g"]


		test_cases = (
				ResolverPlaygroundTestCase(
					["sys-libs/x", "sys-apps/b"],
					options = {"--rebuild-if-unbuilt" : True,
						"--rebuild-exclude" : ["sys-apps/c"]},
					mergelist = ['sys-libs/x-2', 'sys-apps/a-2', 'sys-apps/b-2',
						'sys-apps/e-2', 'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["sys-libs/x", "sys-apps/b"],
					options = {"--rebuild-if-unbuilt" : True},
					mergelist = ['sys-libs/x-2', 'sys-apps/a-2', 'sys-apps/b-2',
						'sys-apps/c-2', 'sys-apps/e-2', 'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["sys-libs/x"],
					options = {"--rebuild-if-unbuilt" : True,
						"--rebuild-ignore" : ["sys-libs/x"]},
					mergelist = ['sys-libs/x-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["sys-libs/x", "sys-apps/b"],
					options = {"--rebuild-if-unbuilt" : True,
						"--rebuild-ignore" : ["sys-apps/b"]},
					mergelist = ['sys-libs/x-2', 'sys-apps/a-2', 'sys-apps/b-2',
						'sys-apps/c-2', 'sys-apps/e-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["=sys-libs/x-1-r1", "sys-apps/b"],
					options = {"--rebuild-if-unbuilt" : True},
					mergelist = ['sys-libs/x-1-r1', 'sys-apps/a-2',
						'sys-apps/b-2', 'sys-apps/c-2', 'sys-apps/e-2',
						'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["=sys-libs/x-1-r1", "sys-apps/b"],
					options = {"--rebuild-if-new-rev" : True},
					mergelist = ['sys-libs/x-1-r1', 'sys-apps/a-2',
						'sys-apps/b-2', 'sys-apps/c-2', 'sys-apps/e-2',
						'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["=sys-libs/x-1-r1"],
					options = {"--rebuild-if-new-ver" : True},
					mergelist = ['sys-libs/x-1-r1'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["sys-libs/x", "sys-apps/b"],
					options = {"--rebuild-if-new-ver" : True},
					mergelist = ['sys-libs/x-2', 'sys-apps/a-2',
						'sys-apps/b-2', 'sys-apps/c-2', 'sys-apps/e-2',
						'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["=sys-libs/x-1"],
					options = {"--rebuild-if-new-rev" : True},
					mergelist = ['sys-libs/x-1'],
					ignore_mergelist_order = True,
					success = True),

				ResolverPlaygroundTestCase(
					["=sys-libs/x-1", "=sys-apps/b-1"],
					options = {"--rebuild-if-unbuilt" : True},
					mergelist = ['sys-libs/x-1', 'sys-apps/a-2',
						'sys-apps/b-1', 'sys-apps/c-2', 'sys-apps/e-2',
						'sys-apps/g-2'],
					ignore_mergelist_order = True,
					success = True),
			)

		playground = ResolverPlayground(ebuilds=ebuilds,
			installed=installed, world=world)

		try:
			for test_case in test_cases:
				playground.run_TestCase(test_case)
				self.assertEqual(test_case.test_success, True, test_case.fail_msg)
		finally:
			playground.cleanup()



# data.py -- Calculated/Discovered Data Values
# Copyright 1998-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import grp
import os
import platform
import pwd

import portage
portage.proxy.lazyimport.lazyimport(globals(),
	'portage.output:colorize',
	'portage.util:writemsg',
	'portage.util.path:first_existing',
	'subprocess'
)
from portage.localization import _

ostype = platform.system()
userland = None
if ostype == "DragonFly" or ostype.endswith("BSD"):
	userland = "BSD"
else:
	userland = "GNU"

lchown = getattr(os, "lchown", None)

if not lchown:
	if ostype == "Darwin":
		def lchown(*_args, **_kwargs):
			pass
	else:
		def lchown(*_args, **_kwargs):
			writemsg(colorize("BAD", "!!!") + _(
				" It seems that os.lchown does not"
				" exist.  Please rebuild python.\n"), noiselevel=-1)
		lchown()

lchown = portage._unicode_func_wrapper(lchown)

def _target_eprefix():
	"""
	Calculate the target EPREFIX, which may be different from
	portage.const.EPREFIX due to cross-prefix support. The result
	is equivalent to portage.settings["EPREFIX"], but the calculation
	is done without the expense of instantiating portage.settings.
	@rtype: str
	@return: the target EPREFIX
	"""
	eprefix = os.environ.get("EPREFIX", portage.const.EPREFIX)
	if eprefix:
		eprefix = portage.util.normalize_path(eprefix)
	return eprefix

def _target_root():
	"""
	Calculate the target ROOT. The result is equivalent to
	portage.settings["ROOT"], but the calculation
	is done without the expense of instantiating portage.settings.
	@rtype: str
	@return: the target ROOT (always ends with a slash)
	"""
	root = os.environ.get("ROOT")
	if not root:
		# Handle either empty or unset ROOT.
		root = os.sep
	root = portage.util.normalize_path(root)
	return root.rstrip(os.sep) + os.sep

def portage_group_warning():
	warn_prefix = colorize("BAD", "*** WARNING ***  ")
	mylines = [
		"For security reasons, only system administrators should be",
		"allowed in the portage group.  Untrusted users or processes",
		"can potentially exploit the portage group for attacks such as",
		"local privilege escalation."
	]
	for x in mylines:
		writemsg(warn_prefix, noiselevel=-1)
		writemsg(x, noiselevel=-1)
		writemsg("\n", noiselevel=-1)
	writemsg("\n", noiselevel=-1)

# Portage has 3 security levels that depend on the uid and gid of the main
# process and are assigned according to the following table:
#
# Privileges  secpass  uid    gid
# normal      0        any    any
# group       1        any    portage_gid
# super       2        0      any
#
# If the "wheel" group does not exist then wheelgid falls back to 0.
# If the "portage" group does not exist then portage_uid falls back to wheelgid.

# If the current user is not root, but has write access to the
# EROOT directory (not due to the 0002 bit), then use "unprivileged"
# mode which sets secpass = 2 and uses the UID and GID of the EROOT
# directory to generate default PORTAGE_INST_GID, PORTAGE_INST_UID,
# PORTAGE_USERNAME, and PORTAGE_GRPNAME settings.
def _unprivileged_mode(eroot, eroot_st):
	return os.getuid() != 0 and os.access(eroot, os.W_OK) and \
		not eroot_st.st_mode & 0o0002

uid = os.getuid()
wheelgid = 0
try:
	wheelgid = grp.getgrnam("wheel")[2]
except KeyError:
	pass

# The portage_uid and portage_gid global constants, and others that
# depend on them are initialized lazily, in order to allow configuration
# via make.conf. Eventually, these constants may be deprecated in favor
# of config attributes, since it's conceivable that multiple
# configurations with different constants could be used simultaneously.
_initialized_globals = set()

def _get_global(k):
	if k in _initialized_globals:
		return globals()[k]

	if k == 'secpass':

		unprivileged = False
		if hasattr(portage, 'settings'):
			unprivileged = "unprivileged" in portage.settings.features
		else:
			# The config class has equivalent code, but we also need to
			# do it here if _disable_legacy_globals() has been called.
			eroot_or_parent = first_existing(os.path.join(
				_target_root(), _target_eprefix().lstrip(os.sep)))
			try:
				eroot_st = os.stat(eroot_or_parent)
			except OSError:
				pass
			else:
				unprivileged = _unprivileged_mode(
					eroot_or_parent, eroot_st)

		v = 0
		if uid == 0:
			v = 2
		elif unprivileged:
			v = 2
		elif _get_global('portage_gid') in os.getgroups():
			v = 1

	elif k in ('portage_gid', 'portage_uid'):

		#Discover the uid and gid of the portage user/group
		keyerror = False
		try:
			portage_uid = pwd.getpwnam(_get_global('_portage_username')).pw_uid
		except KeyError:
			keyerror = True
			portage_uid = 0

		try:
			portage_gid = grp.getgrnam(_get_global('_portage_grpname')).gr_gid
		except KeyError:
			keyerror = True
			portage_gid = 0

		# Suppress this error message if both PORTAGE_GRPNAME and
		# PORTAGE_USERNAME are set to "root", for things like
		# Android (see bug #454060).
		if keyerror and not (_get_global('_portage_username') == "root" and
			_get_global('_portage_grpname') == "root"):
			writemsg(colorize("BAD",
				_("portage: 'portage' user or group missing.")) + "\n", noiselevel=-1)
			writemsg(_(
				"         For the defaults, line 1 goes into passwd, "
				"and 2 into group.\n"), noiselevel=-1)
			writemsg(colorize("GOOD",
				"         portage:x:250:250:portage:/var/tmp/portage:/bin/false") \
				+ "\n", noiselevel=-1)
			writemsg(colorize("GOOD", "         portage::250:portage") + "\n",
				noiselevel=-1)
			portage_group_warning()

		globals()['portage_gid'] = portage_gid
		_initialized_globals.add('portage_gid')
		globals()['portage_uid'] = portage_uid
		_initialized_globals.add('portage_uid')

		if k == 'portage_gid':
			return portage_gid
		if k == 'portage_uid':
			return portage_uid
		raise AssertionError('unknown name: %s' % k)

	elif k == 'userpriv_groups':
		v = [_get_global('portage_gid')]
		if secpass >= 2:
			# Get a list of group IDs for the portage user. Do not use
			# grp.getgrall() since it is known to trigger spurious
			# SIGPIPE problems with nss_ldap.
			cmd = ["id", "-G", _portage_username]

			encoding = portage._encodings['content']
			cmd = [portage._unicode_encode(x,
				encoding=encoding, errors='strict') for x in cmd]
			proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
				stderr=subprocess.STDOUT)
			myoutput = proc.communicate()[0]
			status = proc.wait()
			if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:
				for x in portage._unicode_decode(myoutput,
					encoding=encoding, errors='strict').split():
					try:
						v.append(int(x))
					except ValueError:
						pass
				v = sorted(set(v))

	# Avoid instantiating portage.settings when the desired
	# variable is set in os.environ.
	elif k in ('_portage_grpname', '_portage_username'):
		v = None
		if k == '_portage_grpname':
			env_key = 'PORTAGE_GRPNAME'
		else:
			env_key = 'PORTAGE_USERNAME'

		if env_key in os.environ:
			v = os.environ[env_key]
		elif hasattr(portage, 'settings'):
			v = portage.settings.get(env_key)
		else:
			# The config class has equivalent code, but we also need to
			# do it here if _disable_legacy_globals() has been called.
			eroot_or_parent = first_existing(os.path.join(
				_target_root(), _target_eprefix().lstrip(os.sep)))
			try:
				eroot_st = os.stat(eroot_or_parent)
			except OSError:
				pass
			else:
				if _unprivileged_mode(eroot_or_parent, eroot_st):
					if k == '_portage_grpname':
						try:
							grp_struct = grp.getgrgid(eroot_st.st_gid)
						except KeyError:
							pass
						else:
							v = grp_struct.gr_name
					else:
						try:
							pwd_struct = pwd.getpwuid(eroot_st.st_uid)
						except KeyError:
							pass
						else:
							v = pwd_struct.pw_name

		if v is None:
			v = 'portage'
	else:
		raise AssertionError('unknown name: %s' % k)

	globals()[k] = v
	_initialized_globals.add(k)
	return v

class _GlobalProxy(portage.proxy.objectproxy.ObjectProxy):

	__slots__ = ('_name',)

	def __init__(self, name):
		portage.proxy.objectproxy.ObjectProxy.__init__(self)
		object.__setattr__(self, '_name', name)

	def _get_target(self):
		return _get_global(object.__getattribute__(self, '_name'))

for k in ('portage_gid', 'portage_uid', 'secpass', 'userpriv_groups',
	'_portage_grpname', '_portage_username'):
	globals()[k] = _GlobalProxy(k)
del k

def _init(settings):
	"""
	Use config variables like PORTAGE_GRPNAME and PORTAGE_USERNAME to
	initialize global variables. This allows settings to come from make.conf
	instead of requiring them to be set in the calling environment.
	"""
	if '_portage_grpname' not in _initialized_globals and \
		'_portage_username' not in _initialized_globals:

		# Prevents "TypeError: expected string" errors
		# from grp.getgrnam() with PyPy
		native_string = platform.python_implementation() == 'PyPy'

		v = settings.get('PORTAGE_GRPNAME', 'portage')
		if native_string:
			v = portage._native_string(v)
		globals()['_portage_grpname'] = v
		_initialized_globals.add('_portage_grpname')

		v = settings.get('PORTAGE_USERNAME', 'portage')
		if native_string:
			v = portage._native_string(v)
		globals()['_portage_username'] = v
		_initialized_globals.add('_portage_username')

	if 'secpass' not in _initialized_globals:
		v = 0
		if uid == 0:
			v = 2
		elif "unprivileged" in settings.features:
			v = 2
		elif portage_gid in os.getgroups():
			v = 1
		globals()['secpass'] = v
		_initialized_globals.add('secpass')



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import shutil
import tempfile

from portage import os
from portage.tests import TestCase
from portage.checksum import perform_md5
from portage.util.file_copy import copyfile


class CopyFileTestCase(TestCase):

	def testCopyFile(self):

		tempdir = tempfile.mkdtemp()
		try:
			src_path = os.path.join(tempdir, 'src')
			dest_path = os.path.join(tempdir, 'dest')
			content = b'foo'

			with open(src_path, 'wb') as f:
				f.write(content)

			copyfile(src_path, dest_path)

			self.assertEqual(perform_md5(src_path), perform_md5(dest_path))
		finally:
			shutil.rmtree(tempdir)


class CopyFileSparseTestCase(TestCase):

	def testCopyFileSparse(self):

		tempdir = tempfile.mkdtemp()
		try:
			src_path = os.path.join(tempdir, 'src')
			dest_path = os.path.join(tempdir, 'dest')
			content = b'foo'

			# Use seek to create some sparse blocks. Don't make these
			# files too big, in case the filesystem doesn't support
			# sparse files.
			with open(src_path, 'wb') as f:
				f.write(content)
				f.seek(2**17, 1)
				f.write(content)
				f.seek(2**18, 1)
				f.write(content)
				# Test that sparse blocks are handled correctly at
				# the end of the file (involves seek and truncate).
				f.seek(2**17, 1)

			copyfile(src_path, dest_path)

			self.assertEqual(perform_md5(src_path), perform_md5(dest_path))

			# This last part of the test is expected to fail when sparse
			# copy is not implemented, so set the todo flag in order
			# to tolerate failures.
			self.todo = True

			# If sparse blocks were preserved, then both files should
			# consume the same number of blocks.
			self.assertEqual(
				os.stat(src_path).st_blocks,
				os.stat(dest_path).st_blocks)
		finally:
			shutil.rmtree(tempdir)



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


from portage.dep.soname.SonameAtom import SonameAtom
from portage.tests import TestCase
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor


class TestSonameAtomPickle(TestCase):

	_ALL_PROVIDES = frozenset([SonameAtom('x86_64', 'libc.so.6')])

	def test_soname_atom_pickle(self):
		loop = asyncio._wrap_loop()
		with ForkExecutor(loop=loop) as executor:
			result = loop.run_until_complete(loop.run_in_executor(executor, self._get_all_provides))
		self.assertEqual(self._ALL_PROVIDES, result)

	@classmethod
	def _get_all_provides(cls):
		return cls._ALL_PROVIDES



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (ResolverPlayground,
	ResolverPlaygroundTestCase)

class AutounmaskMultilibUseTestCase(TestCase):

	def testAutounmaskMultilibUse(self):

		self.todo = True

		ebuilds = {
			"x11-proto/xextproto-7.2.1-r1": {
				"EAPI": "5",
				"IUSE": "abi_x86_32 abi_x86_64",
			},
			"x11-libs/libXaw-1.0.11-r2": {
				"EAPI": "5",
				"IUSE": "abi_x86_32 abi_x86_64",
				"RDEPEND": "x11-proto/xextproto[abi_x86_32(-)?,abi_x86_64(-)?]"
			},
			"app-emulation/emul-linux-x86-xlibs-20130224-r2": {
				"EAPI": "5",
				"RDEPEND": "x11-libs/libXaw[abi_x86_32]"
			},
			"games-util/steam-client-meta-0-r20130514": {
				"EAPI": "5",
				"RDEPEND": "app-emulation/emul-linux-x86-xlibs"
			}
		}

		installed = {
			"x11-proto/xextproto-7.2.1-r1": {
				"EAPI": "5",
				"IUSE": "abi_x86_32 abi_x86_64",
				"USE": "abi_x86_32 abi_x86_64"
			},
			"x11-libs/libXaw-1.0.11-r2": {
				"EAPI": "5",
				"IUSE": "abi_x86_32 abi_x86_64",
				"RDEPEND": "x11-proto/xextproto[abi_x86_32(-)?,abi_x86_64(-)?]",
				"USE": "abi_x86_32 abi_x86_64"
			},
			"app-emulation/emul-linux-x86-xlibs-20130224-r2": {
				"EAPI": "5",
				"RDEPEND": "x11-libs/libXaw[abi_x86_32]"
			},
			"games-util/steam-client-meta-0-r20130514": {
				"EAPI": "5",
				"RDEPEND": "app-emulation/emul-linux-x86-xlibs"
			}
		}

		user_config = {
			#"make.conf" : ("USE=\"abi_x86_32 abi_x86_64\"",)
			"make.conf" : ("USE=\"abi_x86_64\"",)
		}

		world = ("games-util/steam-client-meta",)

		test_cases = (

				# Test autounmask solving of multilib use deps for bug #481628.
				# We would like it to suggest some USE changes, but instead it
				# currently fails with a SLOT conflict.

				ResolverPlaygroundTestCase(
					["x11-proto/xextproto", "x11-libs/libXaw"],
					options = {"--oneshot": True, "--autounmask": True,
						"--backtrack": 30},
					mergelist = ["x11-proto/xextproto-7.2.1-r1", "x11-libs/libXaw-1.0.11-r2"],
					success = True),
			)

		playground = ResolverPlayground(ebuilds=ebuilds, installed=installed,
			user_config=user_config, world=world, debug=False)

		try:
			for test_case in test_cases:
				playground.run_TestCase(test_case)
				self.assertEqual(test_case.test_success, True, test_case.fail_msg)
		finally:
			playground.cleanup()



# elog/__init__.py - elog core functions
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys

import portage
portage.proxy.lazyimport.lazyimport(globals(),
	'portage.util:writemsg',
)

from portage.const import EBUILD_PHASES
from portage.exception import AlarmSignal, PortageException
from portage.process import atexit_register
from portage.elog.messages import collect_ebuild_messages, collect_messages
from portage.elog.filtering import filter_loglevels
from portage.localization import _
from portage import os

def _preload_elog_modules(settings):
	logsystems = settings.get("PORTAGE_ELOG_SYSTEM", "").split()
	for s in logsystems:
		# allow per module overrides of PORTAGE_ELOG_CLASSES
		if ":" in s:
			s, levels = s.split(":", 1)
			levels = levels.split(",")
		# - is nicer than _ for module names, so allow people to use it.
		s = s.replace("-", "_")
		try:
			_load_mod("portage.elog.mod_" + s)
		except ImportError:
			pass

def _merge_logentries(a, b):
	rValue = {}
	phases = set(a)
	phases.update(b)
	for p in phases:
		merged_msgs = []
		rValue[p] = merged_msgs
		for d in a, b:
			msgs = d.get(p)
			if msgs:
				merged_msgs.extend(msgs)
	return rValue

def _combine_logentries(logentries):
	# generate a single string with all log messages
	rValue = []
	for phase in EBUILD_PHASES:
		if not phase in logentries:
			continue
		previous_type = None
		for msgtype, msgcontent in logentries[phase]:
			if previous_type != msgtype:
				previous_type = msgtype
				rValue.append("%s: %s" % (msgtype, phase))
			if isinstance(msgcontent, str):
				rValue.append(msgcontent.rstrip("\n"))
			else:
				for line in msgcontent:
					rValue.append(line.rstrip("\n"))
	if rValue:
		rValue.append("")
	return "\n".join(rValue)

_elog_mod_imports = {}
def _load_mod(name):
	global _elog_mod_imports
	m = _elog_mod_imports.get(name)
	if m is None:
		m = __import__(name)
		for comp in name.split(".")[1:]:
			m = getattr(m, comp)
		_elog_mod_imports[name] = m
	return m

_elog_listeners = []
def add_listener(listener):
	'''
	Listeners should accept four arguments: settings, key, logentries and logtext
	'''
	_elog_listeners.append(listener)

def remove_listener(listener):
	'''
	Remove previously added listener
	'''
	_elog_listeners.remove(listener)

_elog_atexit_handlers = []

def elog_process(cpv, mysettings, phasefilter=None):
	global _elog_atexit_handlers

	logsystems = mysettings.get("PORTAGE_ELOG_SYSTEM","").split()
	for s in logsystems:
		# allow per module overrides of PORTAGE_ELOG_CLASSES
		if ":" in s:
			s, levels = s.split(":", 1)
			levels = levels.split(",")
		# - is nicer than _ for module names, so allow people to use it.
		s = s.replace("-", "_")
		try:
			_load_mod("portage.elog.mod_" + s)
		except ImportError:
			pass

	if "T" in mysettings:
		ebuild_logentries = collect_ebuild_messages(
			os.path.join(mysettings["T"], "logging"))
	else:
		# A build dir isn't necessarily required since the messages.e*
		# functions allow messages to be generated in-memory.
		ebuild_logentries = {}
	all_logentries = collect_messages(key=cpv, phasefilter=phasefilter)
	if cpv in all_logentries:
		# Messages generated by the python elog implementation are assumed
		# to come first. For example, this ensures correct order for einfo
		# messages that are generated prior to the setup phase.
		all_logentries[cpv] = \
			_merge_logentries(all_logentries[cpv], ebuild_logentries)
	else:
		all_logentries[cpv] = ebuild_logentries

	my_elog_classes = set(mysettings.get("PORTAGE_ELOG_CLASSES", "").split())
	logsystems = {}
	for token in mysettings.get("PORTAGE_ELOG_SYSTEM", "").split():
		if ":" in token:
			s, levels = token.split(":", 1)
			levels = levels.split(",")
		else:
			s = token
			levels = ()
		levels_set = logsystems.get(s)
		if levels_set is None:
			levels_set = set()
			logsystems[s] = levels_set
		levels_set.update(levels)

	for key in all_logentries:
		default_logentries = filter_loglevels(all_logentries[key], my_elog_classes)

		# in case the filters matched all messages and no module overrides exist
		if len(default_logentries) == 0 and (not ":" in mysettings.get("PORTAGE_ELOG_SYSTEM", "")):
			continue

		default_fulllog = _combine_logentries(default_logentries)

		# call listeners
		for listener in _elog_listeners:
			listener(mysettings, str(key), default_logentries, default_fulllog)

		# pass the processing to the individual modules
		for s, levels in logsystems.items():
			# allow per module overrides of PORTAGE_ELOG_CLASSES
			if levels:
				mod_logentries = filter_loglevels(all_logentries[key], levels)
				mod_fulllog = _combine_logentries(mod_logentries)
			else:
				mod_logentries = default_logentries
				mod_fulllog = default_fulllog
			if len(mod_logentries) == 0:
				continue
			# - is nicer than _ for module names, so allow people to use it.
			s = s.replace("-", "_")
			try:
				m = _load_mod("portage.elog.mod_" + s)
				# Timeout after one minute (in case something like the mail
				# module gets hung).
				try:
					AlarmSignal.register(60)
					m.process(mysettings, str(key), mod_logentries, mod_fulllog)
				finally:
					AlarmSignal.unregister()
				if hasattr(m, "finalize") and not m.finalize in _elog_atexit_handlers:
					_elog_atexit_handlers.append(m.finalize)
					atexit_register(m.finalize)
			except (ImportError, AttributeError) as e:
				writemsg(_("!!! Error while importing logging modules "
					"while loading \"mod_%s\":\n") % str(s))
				writemsg("%s\n" % str(e), noiselevel=-1)
			except AlarmSignal:
				writemsg("Timeout in elog_process for system '%s'\n" % s,
					noiselevel=-1)
			except PortageException as e:
				writemsg("%s\n" % str(e), noiselevel=-1)
